import unittest
from fontParts.base.deprecated import RemovedError


class TestDeprecated(unittest.TestCase):
    # ----
    # Font
    # ----

    def getFont_glyphs(self):
        font, _ = self.objectGenerator("font")
        for name in "ABCD":
            font.newGlyph(name)
        return font

    def test_font_removed_getParent(self):
        font, _ = self.objectGenerator("font")
        with self.assertRaises(RemovedError):
            font.getParent()

    def test_font_removed_generateGlyph(self):
        font, _ = self.objectGenerator("font")
        with self.assertRaises(RemovedError):
            font.generateGlyph()

    def test_font_removed_compileGlyph(self):
        font, _ = self.objectGenerator("font")
        with self.assertRaises(RemovedError):
            font.compileGlyph()

    def test_font_removed_getGlyphNameToFileNameFunc(self):
        font, _ = self.objectGenerator("font")
        with self.assertRaises(RemovedError):
            font.getGlyphNameToFileNameFunc()

    def test_font_deprecated_update(self):
        # As changed() is defined by the environment, only test if a Warning is issued.
        font, _ = self.objectGenerator("font")
        with self.assertWarnsRegex(DeprecationWarning, "Font.changed()"):
            font.update()

    def test_font_deprecated_setChanged(self):
        # As changed() is defined by the environment, only test if a Warning is issued.
        font, _ = self.objectGenerator("font")
        with self.assertWarnsRegex(DeprecationWarning, "Font.changed()"):
            font.setChanged()

    def test_font_removed_setParent(self):
        font, _ = self.objectGenerator("font")
        with self.assertRaises(RemovedError):
            font.setParent(font)

    def test_font_deprecated__fileName(self):
        font, _ = self.objectGenerator("font")
        with self.assertWarnsRegex(DeprecationWarning, "fileName"):
            font._get_fileName()
        self.assertEqual(font._get_fileName(), font.path)

    def test_font_deprecated_fileName(self):
        font, _ = self.objectGenerator("font")
        with self.assertWarnsRegex(DeprecationWarning, "fileName"):
            font.fileName
        self.assertEqual(font.fileName, font.path)

    def test_font_deprecated_getWidth(self):
        font, _ = self.objectGenerator("font")
        glyph = font.newGlyph("Test")
        glyph.width = 200
        with self.assertWarnsRegex(DeprecationWarning, "Font.getWidth()"):
            font.getWidth("Test")
        self.assertEqual(font.getWidth("Test"), font["Test"].width)

    def test_font_deprecated_getGlyph(self):
        font, _ = self.objectGenerator("font")
        font.newGlyph("Test")
        with self.assertWarnsRegex(DeprecationWarning, "Font.getGlyph()"):
            font.getGlyph("Test")
        self.assertEqual(font.getGlyph("Test"), font["Test"])

    def test_font_deprecated__get_selection(self):
        font = self.getFont_glyphs()
        try:
            font.defaultLayer.selected = False
        except NotImplementedError:
            return
        glyph1 = font["A"]
        glyph2 = font["B"]
        glyph1.selected = True
        glyph2.selected = True
        with self.assertWarnsRegex(DeprecationWarning, "Font.selectedGlyphNames"):
            font._get_selection()
        self.assertEqual(font._get_selection(), font.selectedGlyphNames)

    def test_font_deprecated__set_selection(self):
        font1 = self.getFont_glyphs()
        font2 = self.getFont_glyphs()
        with self.assertWarnsRegex(DeprecationWarning, "Font.selectedGlyphNames"):
            font1._set_selection(["A", "B"])
        font2.selectedGlyphNames = ["A", "B"]
        self.assertEqual(font1.selectedGlyphNames, font2.selectedGlyphNames)

    def test_font_deprecated_selection_set(self):
        font1 = self.getFont_glyphs()
        font2 = self.getFont_glyphs()
        with self.assertWarnsRegex(DeprecationWarning, "Font.selectedGlyphNames"):
            font1.selection = ["A", "B"]
        font2.selectedGlyphNames = ["A", "B"]
        self.assertEqual(font1.selectedGlyphNames, font2.selectedGlyphNames)

    def test_font_deprecated_selection_get(self):
        font = self.getFont_glyphs()
        try:
            font.defaultLayer.selected = False
        except NotImplementedError:
            return
        glyph1 = font["A"]
        glyph2 = font["B"]
        glyph1.selected = True
        glyph2.selected = True
        with self.assertWarnsRegex(DeprecationWarning, "Font.selectedGlyphNames"):
            font.selection
        self.assertEqual(font.selection, font.selectedGlyphNames)

    # ------
    # Anchor
    # ------

    def getAnchor(self):
        glyph, _ = self.objectGenerator("glyph")
        glyph.appendAnchor("anchor 0", (10, 20))
        return glyph.anchors[0]

    def test_anchor_deprecated__generateIdentifer(self):
        anchor, _ = self.objectGenerator("anchor")
        with self.assertWarnsRegex(DeprecationWarning, "Anchor._generateIdentifier()"):
            anchor._generateIdentifier()
        self.assertEqual(anchor._generateIdentifier(), anchor._getIdentifier())

    def test_anchor_deprecated_generateIdentifer(self):
        anchor, _ = self.objectGenerator("anchor")
        with self.assertWarnsRegex(DeprecationWarning, "Anchor.generateIdentifier()"):
            anchor.generateIdentifier()
        self.assertEqual(anchor.generateIdentifier(), anchor.getIdentifier())

    def test_anchor_deprecated_getParent(self):
        anchor = self.getAnchor()
        with self.assertWarnsRegex(DeprecationWarning, "Anchor.getParent()"):
            anchor.getParent()
        self.assertEqual(anchor.getParent(), anchor.glyph)

    def test_anchor_deprecated_update(self):
        # As changed() is defined by the environment, only test if a Warning is issued.
        anchor, _ = self.objectGenerator("anchor")
        with self.assertWarnsRegex(DeprecationWarning, "Anchor.changed()"):
            anchor.update()

    def test_anchor_deprecated_setChanged(self):
        # As changed() is defined by the environment, only test if a Warning is issued.
        anchor, _ = self.objectGenerator("anchor")
        with self.assertWarnsRegex(DeprecationWarning, "Anchor.changed()"):
            anchor.setChanged()

    def test_anchor_removed_setParent(self):
        anchor = self.getAnchor()
        glyph = anchor.glyph
        with self.assertRaises(RemovedError):
            anchor.setParent(glyph)

    def test_anchor_removed_draw(self):
        anchor = self.getAnchor()
        pen = anchor.glyph.getPen()
        with self.assertRaises(RemovedError):
            anchor.draw(pen)

    def test_anchor_removed_drawPoints(self):
        anchor = self.getAnchor()
        pen = anchor.glyph.getPen()
        with self.assertRaises(RemovedError):
            anchor.drawPoints(pen)

    def test_anchor_deprecated_move(self):
        anchor1 = self.getAnchor()
        anchor2 = self.getAnchor()
        with self.assertWarnsRegex(DeprecationWarning, "Anchor.move()"):
            anchor1.move((0, 20))
        self.assertNotEqual((anchor1.x, anchor1.y), (anchor2.x, anchor2.y))
        anchor2.moveBy((0, 20))
        self.assertEqual((anchor1.x, anchor1.y), (anchor2.x, anchor2.y))

    def test_anchor_deprecated_translate(self):
        anchor1 = self.getAnchor()
        anchor2 = self.getAnchor()
        with self.assertWarnsRegex(DeprecationWarning, "Anchor.translate()"):
            anchor1.translate((0, 20))
        self.assertNotEqual((anchor1.x, anchor1.y), (anchor2.x, anchor2.y))
        anchor2.moveBy((0, 20))
        self.assertEqual((anchor1.x, anchor1.y), (anchor2.x, anchor2.y))

    def test_anchor_deprecated_scale_no_center(self):
        anchor1 = self.getAnchor()
        anchor2 = self.getAnchor()
        with self.assertWarnsRegex(DeprecationWarning, "Anchor.scale()"):
            anchor1.scale(-2)
        self.assertNotEqual((anchor1.x, anchor1.y), (anchor2.x, anchor2.y))
        anchor2.scaleBy(-2)
        self.assertEqual((anchor1.x, anchor1.y), (anchor2.x, anchor2.y))

    def test_anchor_deprecated_scale_center(self):
        anchor1 = self.getAnchor()
        anchor2 = self.getAnchor()
        with self.assertWarnsRegex(DeprecationWarning, "Anchor.scale()"):
            anchor1.scale((-2, 3), center=(1, 2))
        self.assertNotEqual((anchor1.x, anchor1.y), (anchor2.x, anchor2.y))
        anchor2.scaleBy((-2, 3), origin=(1, 2))
        self.assertEqual((anchor1.x, anchor1.y), (anchor2.x, anchor2.y))

    def test_anchor_deprecated_rotate_no_offset(self):
        anchor1 = self.getAnchor()
        anchor2 = self.getAnchor()
        with self.assertWarnsRegex(DeprecationWarning, "Anchor.rotate()"):
            anchor1.rotate(45)
        self.assertNotEqual((anchor1.x, anchor1.y), (anchor2.x, anchor2.y))
        anchor2.rotateBy(45)
        self.assertEqual((anchor1.x, anchor1.y), (anchor2.x, anchor2.y))

    def test_anchor_deprecated_rotate_offset(self):
        anchor1 = self.getAnchor()
        anchor2 = self.getAnchor()
        with self.assertWarnsRegex(DeprecationWarning, "Anchor.rotate()"):
            anchor1.rotate(45, offset=(1, 2))
        self.assertNotEqual((anchor1.x, anchor1.y), (anchor2.x, anchor2.y))
        anchor2.rotateBy(45, origin=(1, 2))
        self.assertEqual((anchor1.x, anchor1.y), (anchor2.x, anchor2.y))

    def test_anchor_deprecated_transform(self):
        anchor1 = self.getAnchor()
        anchor2 = self.getAnchor()
        with self.assertWarnsRegex(DeprecationWarning, "Anchor.transform()"):
            anchor1.transform((2, 0, 0, 3, -3, 2))
        self.assertNotEqual((anchor1.x, anchor1.y), (anchor2.x, anchor2.y))
        anchor2.transformBy((2, 0, 0, 3, -3, 2))
        self.assertEqual((anchor1.x, anchor1.y), (anchor2.x, anchor2.y))

    def test_anchor_deprecated_skew_no_offset_one_value(self):
        anchor1 = self.getAnchor()
        anchor2 = self.getAnchor()
        with self.assertWarnsRegex(DeprecationWarning, "Anchor.skew()"):
            anchor1.skew(100)
        self.assertNotEqual((anchor1.x, anchor1.y), (anchor2.x, anchor2.y))
        anchor2.skewBy(100)
        self.assertEqual((anchor1.x, anchor1.y), (anchor2.x, anchor2.y))

    def test_anchor_deprecated_skew_no_offset_two_values(self):
        anchor1 = self.getAnchor()
        anchor2 = self.getAnchor()
        with self.assertWarnsRegex(DeprecationWarning, "Anchor.skew()"):
            anchor1.skew((100, 200))
        self.assertNotEqual((anchor1.x, anchor1.y), (anchor2.x, anchor2.y))
        anchor2.skewBy((100, 200))
        self.assertEqual((anchor1.x, anchor1.y), (anchor2.x, anchor2.y))

    def test_anchor_deprecated_skew_offset_one_value(self):
        anchor1 = self.getAnchor()
        anchor2 = self.getAnchor()
        with self.assertWarnsRegex(DeprecationWarning, "Anchor.skew()"):
            anchor1.skew(100, offset=(1, 2))
        self.assertNotEqual((anchor1.x, anchor1.y), (anchor2.x, anchor2.y))
        anchor2.skewBy(100, origin=(1, 2))
        self.assertEqual((anchor1.x, anchor1.y), (anchor2.x, anchor2.y))

    def test_anchor_deprecated_skew_offset_two_values(self):
        anchor1 = self.getAnchor()
        anchor2 = self.getAnchor()
        with self.assertWarnsRegex(DeprecationWarning, "Anchor.skew()"):
            anchor1.skew((100, 200), offset=(1, 2))
        self.assertNotEqual((anchor1.x, anchor1.y), (anchor2.x, anchor2.y))
        anchor2.skewBy((100, 200), origin=(1, 2))
        self.assertEqual((anchor1.x, anchor1.y), (anchor2.x, anchor2.y))

    # -----
    # Layer
    # -----

    def test_layer_deprecated_getParent(self):
        font, _ = self.objectGenerator("font")
        for name in "ABCD":
            font.newLayer("layer " + name)
        layer = font.layers[0]
        with self.assertWarnsRegex(DeprecationWarning, "Layer.font"):
            layer.getParent()
        self.assertEqual(layer.getParent(), layer.font)

    def test_layer_deprecated_update(self):
        # As changed() is defined by the environment, only test if a Warning is issued.
        layer, _ = self.objectGenerator("layer")
        with self.assertWarnsRegex(DeprecationWarning, "Layer.changed()"):
            layer.update()

    def test_layer_deprecated_setChanged(self):
        # As changed() is defined by the environment, only test if a Warning is issued.
        layer, _ = self.objectGenerator("layer")
        with self.assertWarnsRegex(DeprecationWarning, "Layer.changed()"):
            layer.setChanged()

    def test_layer_removed_setParent(self):
        font, _ = self.objectGenerator("font")
        for name in "ABCD":
            font.newLayer("layer " + name)
        layer = font.layers[0]
        with self.assertRaises(RemovedError):
            layer.setParent(font)

    # --------
    # Features
    # --------

    def test_features_deprecated_getParent(self):
        font, _ = self.objectGenerator("font")
        features = font.features
        features.text = "# Test"
        with self.assertWarnsRegex(DeprecationWarning, "Features.font"):
            features.getParent()
        self.assertEqual(features.getParent(), features.font)

    def test_features_deprecated_update(self):
        # As changed() is defined by the environment, only test if a Warning is issued.
        features, _ = self.objectGenerator("features")
        with self.assertWarnsRegex(DeprecationWarning, "Features.changed()"):
            features.update()

    def test_features_deprecated_setChanged(self):
        # As changed() is defined by the environment, only test if a Warning is issued.
        features, _ = self.objectGenerator("features")
        with self.assertWarnsRegex(DeprecationWarning, "Features.changed()"):
            features.setChanged()

    def test_feature_removed_setParent(self):
        font, _ = self.objectGenerator("font")
        features = font.features
        features.text = "# Test"
        with self.assertRaises(RemovedError):
            features.setParent(font)

    def test_feature_removed_round(self):
        feature, _ = self.objectGenerator("features")
        with self.assertRaises(RemovedError):
            feature.round()

    # -----
    # Image
    # -----

    def getImage_glyph(self):
        from fontParts.test.test_image import testImageData

        glyph, _ = self.objectGenerator("glyph")
        glyph.addImage(data=testImageData)
        image = glyph.image
        return image

    def test_image_deprecated_getParent(self):
        image = self.getImage_glyph()
        with self.assertWarnsRegex(DeprecationWarning, "Image.glyph"):
            image.getParent()
        self.assertEqual(image.getParent(), image.glyph)

    def test_image_deprecated_update(self):
        # As changed() is defined by the environment, only test if a Warning is issued.
        image, _ = self.objectGenerator("image")
        with self.assertWarnsRegex(DeprecationWarning, "Image.changed()"):
            image.update()

    def test_image_deprecated_setChanged(self):
        # As changed() is defined by the environment, only test if a Warning is issued.
        image, _ = self.objectGenerator("image")
        with self.assertWarnsRegex(DeprecationWarning, "Image.changed()"):
            image.setChanged()

    def test_image_removed_setParent(self):
        glyph, _ = self.objectGenerator("glyph")
        image = self.getImage_glyph()
        with self.assertRaises(RemovedError):
            image.setParent(glyph)

    # ----
    # Info
    # ----

    def test_info_deprecated_getParent(self):
        font, _ = self.objectGenerator("font")
        info = font.info
        info.unitsPerEm = 1000
        with self.assertWarnsRegex(DeprecationWarning, "Info.font"):
            info.getParent()
        self.assertEqual(info.getParent(), info.font)

    def test_info_deprecated_setChanged(self):
        # As changed() is defined by the environment, only test if a Warning is issued.
        info, _ = self.objectGenerator("info")
        with self.assertWarnsRegex(DeprecationWarning, "Info.changed()"):
            info.setChanged()

    def test_info_removed_setParent(self):
        font, _ = self.objectGenerator("font")
        info, _ = self.objectGenerator("info")
        info.unitsPerEm = 1000
        with self.assertRaises(RemovedError):
            info.setParent(font)

    # -------
    # Kerning
    # -------

    def getKerning_generic(self):
        font, _ = self.objectGenerator("font")
        groups = font.groups
        groups["public.kern1.X"] = ["A", "B", "C"]
        groups["public.kern2.X"] = ["A", "B", "C"]
        kerning = font.kerning
        kerning.update(
            {
                ("public.kern1.X", "public.kern2.X"): 100,
                ("B", "public.kern2.X"): 101,
                ("public.kern1.X", "B"): 102,
                ("A", "A"): 103,
            }
        )
        return kerning

    def test_kerning_removed_setParent(self):
        font, _ = self.objectGenerator("font")
        kerning, _ = self.objectGenerator("kerning")
        with self.assertRaises(RemovedError):
            kerning.setParent(font)

    def test_kerning_removed_swapNames(self):
        kerning = self.getKerning_generic()
        swap = {"B": "C"}
        with self.assertRaises(RemovedError):
            kerning.swapNames(swap)

    def test_kerning_removed_getLeft(self):
        kerning = self.getKerning_generic()
        with self.assertRaises(RemovedError):
            kerning.getLeft("B")

    def test_kerning_removed_getRight(self):
        kerning = self.getKerning_generic()
        with self.assertRaises(RemovedError):
            kerning.getRight("B")

    def test_kerning_removed_getExtremes(self):
        kerning = self.getKerning_generic()
        with self.assertRaises(RemovedError):
            kerning.getExtremes()

    def test_kerning_removed_add(self):
        kerning = self.getKerning_generic()
        with self.assertRaises(RemovedError):
            kerning.add(10)

    def test_kerning_removed_minimize(self):
        kerning = self.getKerning_generic()
        with self.assertRaises(RemovedError):
            kerning.minimize()

    def test_kerning_removed_importAFM(self):
        kerning = self.getKerning_generic()
        with self.assertRaises(RemovedError):
            kerning.importAFM("fake/path")

    def test_kerning_removed_getAverage(self):
        kerning = self.getKerning_generic()
        with self.assertRaises(RemovedError):
            kerning.getAverage()

    def test_kerning_removed_combine(self):
        kerning = self.getKerning_generic()
        one = {("A", "v"): -10}
        two = {("v", "A"): -10}
        with self.assertRaises(RemovedError):
            kerning.combine([one, two])

    def test_kerning_removed_eliminate(self):
        kerning = self.getKerning_generic()
        with self.assertRaises(RemovedError):
            kerning.eliminate(leftGlyphsToEliminate=["A"])

    def test_kerning_removed_occurrenceCount(self):
        kerning = self.getKerning_generic()
        with self.assertRaises(RemovedError):
            kerning.occurrenceCount(["A"])

    def test_kerning_removed_implodeClasses(self):
        kerning = self.getKerning_generic()
        classes = {"group": ["a", "v"]}
        with self.assertRaises(RemovedError):
            kerning.implodeClasses(leftClassDict=classes)

    def test_kerning_removed_explodeClasses(self):
        kerning = self.getKerning_generic()
        classes = {"group": ["a", "v"]}
        with self.assertRaises(RemovedError):
            kerning.explodeClasses(leftClassDict=classes)

    def test_kerning_removed_setChanged(self):
        kerning = self.getKerning_generic()
        # As changed() is defined by the environment, only test if a Warning is issued.
        with self.assertWarnsRegex(DeprecationWarning, "Kerning.changed()"):
            kerning.setChanged()

    def test_kerning_removed_getParent(self):
        kerning = self.getKerning_generic()
        with self.assertWarnsRegex(DeprecationWarning, "Kerning.font"):
            kerning.getParent()
        self.assertEqual(kerning.getParent(), kerning.font)

    # ------
    # Groups
    # ------

    def test_groups_deprecated_getParent(self):
        font, _ = self.objectGenerator("font")
        groups = font.groups
        groups.update(
            {
                "group 1": ["A", "B", "C"],
                "group 2": ["x", "y", "z"],
                "group 3": [],
                "group 4": ["A"],
            }
        )
        with self.assertWarnsRegex(DeprecationWarning, "Groups.font"):
            groups.getParent()
        self.assertEqual(groups.getParent(), groups.font)

    def test_groups_deprecated_setChanged(self):
        # As changed() is defined by the environment, only test if a Warning is issued.
        groups, _ = self.objectGenerator("groups")
        with self.assertWarnsRegex(DeprecationWarning, "Groups.changed()"):
            groups.setChanged()

    def test_groups_removed_setParent(self):
        font, _ = self.objectGenerator("font")
        groups, _ = self.objectGenerator("groups")
        groups.update(
            {
                "group 1": ["A", "B", "C"],
                "group 2": ["x", "y", "z"],
                "group 3": [],
                "group 4": ["A"],
            }
        )
        with self.assertRaises(RemovedError):
            groups.setParent(font)

    # ---
    # Lib
    # ---

    def test_lib_deprecated_getParent_font(self):
        font, _ = self.objectGenerator("font")
        lib = font.lib
        lib.update({"key 1": ["A", "B", "C"], "key 2": "x", "key 3": [], "key 4": 20})
        with self.assertWarnsRegex(DeprecationWarning, "Lib.font"):
            lib.getParent()
        self.assertEqual(lib.getParent(), lib.font)

    def test_lib_deprecated_getParent_glyph(self):
        font, _ = self.objectGenerator("font")
        glyph = font.newGlyph("Test")
        lib = glyph.lib
        lib.update({"key 1": ["A", "B", "C"], "key 2": "x", "key 3": [], "key 4": 20})
        with self.assertWarnsRegex(DeprecationWarning, "Lib.glyph"):
            lib.getParent()
        self.assertEqual(lib.getParent(), lib.glyph)

    def test_lib_deprecated_setChanged(self):
        # As changed() is defined by the environment, only test if a Warning is issued.
        lib, _ = self.objectGenerator("lib")
        with self.assertWarnsRegex(DeprecationWarning, "Lib.changed()"):
            lib.setChanged()

    def test_lib_removed_setParent_font(self):
        font, _ = self.objectGenerator("font")
        lib, _ = self.objectGenerator("lib")
        lib.update({"key 1": ["A", "B", "C"], "key 2": "x", "key 3": [], "key 4": 20})
        with self.assertRaises(RemovedError):
            lib.setParent(font)

    def test_lib_removed_setParent_glyph(self):
        glyph, _ = self.objectGenerator("glyph")
        lib, _ = self.objectGenerator("lib")
        lib.update({"key 1": ["A", "B", "C"], "key 2": "x", "key 3": [], "key 4": 20})
        with self.assertRaises(RemovedError):
            lib.setParent(glyph)

    # ---------
    # Guideline
    # ---------

    def getGuideline_generic(self):
        guideline, _ = self.objectGenerator("guideline")
        guideline.x = 1
        guideline.y = 2
        guideline.angle = 90
        return guideline

    def getGuideline_transform(self):
        guideline = self.getGuideline_generic()
        guideline.angle = 45.0
        return guideline

    def test_guideline_deprecated__generateIdentifer(self):
        guideline = self.getGuideline_generic()
        with self.assertWarnsRegex(DeprecationWarning, "Guideline._getIdentifier()"):
            guideline._generateIdentifier()
        self.assertEqual(guideline._generateIdentifier(), guideline._getIdentifier())

    def test_guideline_deprecated_generateIdentifer(self):
        guideline = self.getGuideline_generic()
        with self.assertWarnsRegex(DeprecationWarning, "Guideline.getIdentifier()"):
            guideline.generateIdentifier()
        self.assertEqual(guideline.generateIdentifier(), guideline.getIdentifier())

    def test_guideline_deprecated_getParent_glyph(self):
        glyph, _ = self.objectGenerator("glyph")
        guideline = self.getGuideline_generic()
        guideline.glyph = glyph
        with self.assertWarnsRegex(DeprecationWarning, "Guideline.glyph"):
            guideline.getParent()
        self.assertEqual(guideline.getParent(), guideline.glyph)

    def test_guideline_deprecated_getParent_font(self):
        font, _ = self.objectGenerator("font")
        guideline = self.getGuideline_generic()
        guideline.font = font
        with self.assertWarnsRegex(DeprecationWarning, "Guideline.font"):
            guideline.getParent()
        self.assertEqual(guideline.getParent(), guideline.font)

    def test_guideline_deprecated_update(self):
        # As changed() is defined by the environment, only test if a Warning is issued.
        guideline, _ = self.objectGenerator("guideline")
        with self.assertWarnsRegex(DeprecationWarning, "Guideline.changed()"):
            guideline.update()

    def test_guideline_deprecated_setChanged(self):
        # As changed() is defined by the environment, only test if a Warning is issued.
        guideline, _ = self.objectGenerator("guideline")
        with self.assertWarnsRegex(DeprecationWarning, "Guideline.changed()"):
            guideline.setChanged()

    def test_guideline_removed_setParent(self):
        font, _ = self.objectGenerator("font")
        guideline = self.getGuideline_generic()
        with self.assertRaises(RemovedError):
            guideline.setParent(font)

    def test_guideline_deprecated_move(self):
        guideline1, _ = self.objectGenerator("guideline")
        guideline2, _ = self.objectGenerator("guideline")
        with self.assertWarnsRegex(DeprecationWarning, "Guideline.move()"):
            guideline1.move((0, 20))
        guideline2.moveBy((0, 20))
        self.assertEqual(guideline1.y, guideline2.y)

    def test_guideline_deprecated_translate(self):
        guideline1, _ = self.objectGenerator("guideline")
        guideline2, _ = self.objectGenerator("guideline")
        with self.assertWarnsRegex(DeprecationWarning, "Guideline.translate()"):
            guideline1.translate((0, 20))
        guideline2.moveBy((0, 20))
        self.assertEqual(guideline1.y, guideline2.y)

    def test_guideline_deprecated_scale_no_center(self):
        guideline = self.getGuideline_transform()
        with self.assertWarnsRegex(DeprecationWarning, "Guideline.scale()"):
            guideline.scale(-2)
        self.assertEqual(guideline.x, -2)
        self.assertEqual(guideline.y, -4)
        self.assertAlmostEqual(guideline.angle, 225.000, places=3)

    def test_guideline_deprecated_scale_center(self):
        guideline = self.getGuideline_transform()
        with self.assertWarnsRegex(DeprecationWarning, "Guideline.scale()"):
            guideline.scale((-2, 3), center=(1, 2))
        self.assertEqual(guideline.x, 1)
        self.assertEqual(guideline.y, 2)
        self.assertAlmostEqual(guideline.angle, 123.690, places=3)

    def test_guideline_deprecated_rotate_no_offset(self):
        guideline = self.getGuideline_transform()
        with self.assertWarnsRegex(DeprecationWarning, "Guideline.rotate()"):
            guideline.rotate(45)
        self.assertAlmostEqual(guideline.x, -0.707, places=3)
        self.assertAlmostEqual(guideline.y, 2.121, places=3)
        self.assertAlmostEqual(guideline.angle, 0.000, places=3)

    def test_guideline_deprecated_rotate_offset(self):
        guideline = self.getGuideline_transform()
        with self.assertWarnsRegex(DeprecationWarning, "Guideline.rotate()"):
            guideline.rotate(45, offset=(1, 2))
        self.assertAlmostEqual(guideline.x, 1)
        self.assertAlmostEqual(guideline.y, 2)
        self.assertAlmostEqual(guideline.angle, 0.000, places=3)

    def test_guideline_deprecated_transform(self):
        guideline = self.getGuideline_transform()
        with self.assertWarnsRegex(DeprecationWarning, "Guideline.transform()"):
            guideline.transform((2, 0, 0, 3, -3, 2))
        self.assertEqual(guideline.x, -1)
        self.assertEqual(guideline.y, 8)
        self.assertAlmostEqual(guideline.angle, 56.310, places=3)

    def test_guideline_deprecated_skew_no_offset_one_value(self):
        guideline = self.getGuideline_transform()
        with self.assertWarnsRegex(DeprecationWarning, "Guideline.skew()"):
            guideline.skew(100)
        self.assertAlmostEqual(guideline.x, -10.343, places=3)
        self.assertEqual(guideline.y, 2.0)
        self.assertAlmostEqual(guideline.angle, 8.525, places=3)

    def test_guideline_deprecated_skew_no_offset_two_values(self):
        guideline = self.getGuideline_transform()
        with self.assertWarnsRegex(DeprecationWarning, "Guideline.skew()"):
            guideline.skew((100, 200))
        self.assertAlmostEqual(guideline.x, -10.343, places=3)
        self.assertAlmostEqual(guideline.y, 2.364, places=3)
        self.assertAlmostEqual(guideline.angle, 5.446, places=3)

    def test_guideline_deprecated_skew_offset_one_value(self):
        guideline = self.getGuideline_transform()
        with self.assertWarnsRegex(DeprecationWarning, "Guideline.skew()"):
            guideline.skew(100, offset=(1, 2))
        self.assertEqual(guideline.x, 1)
        self.assertEqual(guideline.y, 2)
        self.assertAlmostEqual(guideline.angle, 8.525, places=3)

    def test_guideline_deprecated_skew_offset_two_values(self):
        guideline = self.getGuideline_transform()
        with self.assertWarnsRegex(DeprecationWarning, "Guideline.skew()"):
            guideline.skew((100, 200), offset=(1, 2))
        self.assertEqual(guideline.x, 1)
        self.assertEqual(guideline.y, 2)
        self.assertAlmostEqual(guideline.angle, 5.446, places=3)

    # -----
    # Glyph
    # -----

    def getGlyph_generic(self):
        glyph, _ = self.objectGenerator("glyph")
        glyph.name = "Test Glyph 1"
        glyph.unicode = int(ord("X"))
        glyph.width = 250
        glyph.height = 750
        pen = glyph.getPen()
        pen.moveTo((100, -10))
        pen.lineTo((100, 100))
        pen.lineTo((200, 100))
        pen.lineTo((200, 0))
        pen.closePath()
        pen.moveTo((110, 10))
        pen.lineTo((110, 90))
        pen.lineTo((190, 90))
        pen.lineTo((190, 10))
        pen.closePath()
        glyph.appendAnchor("Test Anchor 1", (1, 2))
        glyph.appendAnchor("Test Anchor 2", (3, 4))
        glyph.appendGuideline((1, 2), 0, "Test Guideline 1")
        glyph.appendGuideline((3, 4), 90, "Test Guideline 2")
        return glyph

    def test_glyph_removed_center(self):
        glyph = self.getGlyph_generic()
        with self.assertRaisesRegex(RemovedError, "center()"):
            glyph.center()

    def test_glyph_removed_clearVGuides(self):
        glyph = self.getGlyph_generic()
        with self.assertRaisesRegex(RemovedError, "clearGuidelines()"):
            glyph.clearVGuides()

    def test_glyph_removed_clearHGuides(self):
        glyph = self.getGlyph_generic()
        with self.assertRaisesRegex(RemovedError, "clearGuidelines()"):
            glyph.clearHGuides()

    def test_glyph_removed_setParent(self):
        font, _ = self.objectGenerator("font")
        glyph = self.getGlyph_generic()
        with self.assertRaisesRegex(RemovedError, "setParent()"):
            glyph.setParent(font)

    def test_glyph_deprecated_get_mark(self):
        glyph = self.getGlyph_generic()
        glyph.markColor = (1, 0, 0, 1)
        with self.assertWarnsRegex(DeprecationWarning, "Glyph.markColor"):
            glyph._get_mark()
        self.assertEqual(glyph._get_mark(), glyph.markColor)

    def test_glyph_deprecated_set_mark(self):
        glyph = self.getGlyph_generic()
        with self.assertWarnsRegex(DeprecationWarning, "Glyph.markColor"):
            glyph._set_mark((1, 0, 0, 1))
        self.assertEqual((1, 0, 0, 1), glyph.markColor)

    def test_glyph_deprecated_mark(self):
        glyph = self.getGlyph_generic()
        with self.assertWarnsRegex(DeprecationWarning, "Glyph.markColor"):
            glyph.mark = (1, 0, 0, 1)
        with self.assertWarnsRegex(DeprecationWarning, "Glyph.markColor"):
            mark = glyph.mark
        self.assertEqual((1, 0, 0, 1), mark)

    def test_glyph_deprecated__get_box(self):
        glyph = self.getGlyph_generic()
        with self.assertWarnsRegex(DeprecationWarning, "Glyph.bounds"):
            glyph._get_box()
        self.assertEqual(glyph._get_box(), glyph.bounds)

    def test_glyph_deprecated_box(self):
        glyph = self.getGlyph_generic()
        with self.assertWarnsRegex(DeprecationWarning, "Glyph.bounds"):
            box = glyph.box
        self.assertEqual(box, (100, -10, 200, 100))

    def test_glyph_deprecated_getAnchors(self):
        glyph = self.getGlyph_generic()
        with self.assertWarnsRegex(DeprecationWarning, "Glyph.anchors"):
            anchors = glyph.getAnchors()
        self.assertEqual(anchors, glyph.anchors)

    def test_glyph_deprecated_getComponents(self):
        glyph = self.getGlyph_generic()
        glyph.appendComponent("component 1")
        with self.assertWarnsRegex(DeprecationWarning, "Glyph.components"):
            components = glyph.getComponents()
        self.assertEqual(components, glyph.components)

    def test_glyph_deprecated_getParent(self):
        font, _ = self.objectGenerator("font")
        layer = font.layers[0]
        glyph = layer.newGlyph("A")
        with self.assertWarnsRegex(DeprecationWarning, "Glyph.font"):
            parent = glyph.getParent()
        self.assertEqual(parent, glyph.font)

    def test_glyph_deprecated_writeGlyphToString(self):
        glyph = self.getGlyph_generic()
        with self.assertWarnsRegex(DeprecationWarning, "Glyph.dumpToGLIF()"):
            data = glyph.writeGlyphToString()
        self.assertEqual(data, glyph.dumpToGLIF())

    def test_glyph_deprecated_readGlyphToString(self):
        glyph = self.getGlyph_generic()
        glyph2, _ = self.objectGenerator("glyph")
        data = glyph.dumpToGLIF()
        with self.assertWarnsRegex(DeprecationWarning, "Glyph.loadFromGLIF()"):
            glyph2.readGlyphFromString(data)
        self.assertEqual(glyph.bounds, glyph2.bounds)
        self.assertEqual(len(glyph), len(glyph2))

    def test_glyph_deprecated_update(self):
        # As changed() is defined by the environment, only test if a Warning is issued.
        glyph, _ = self.objectGenerator("glyph")
        with self.assertWarnsRegex(DeprecationWarning, "Glyph.changed()"):
            glyph.update()

    def test_glyph_deprecated_setChanged(self):
        # As changed() is defined by the environment, only test if a Warning is issued.
        glyph, _ = self.objectGenerator("glyph")
        with self.assertWarnsRegex(DeprecationWarning, "Glyph.changed()"):
            glyph.setChanged()

    def test_glyph_deprecated_move(self):
        glyph1 = self.getGlyph_generic()
        glyph2 = self.getGlyph_generic()
        with self.assertWarnsRegex(DeprecationWarning, "Glyph.move()"):
            glyph1.move((0, 20))
        self.assertNotEqual(glyph1.bounds, glyph2.bounds)
        glyph2.moveBy((0, 20))
        self.assertEqual(glyph1.bounds, glyph2.bounds)

    def test_glyph_deprecated_translate(self):
        glyph1 = self.getGlyph_generic()
        glyph2 = self.getGlyph_generic()
        with self.assertWarnsRegex(DeprecationWarning, "Glyph.translate()"):
            glyph1.translate((0, 20))
        self.assertNotEqual(glyph1.bounds, glyph2.bounds)
        glyph2.moveBy((0, 20))
        self.assertEqual(glyph1.bounds, glyph2.bounds)

    def test_glyph_deprecated_scale_no_center(self):
        glyph1 = self.getGlyph_generic()
        glyph2 = self.getGlyph_generic()
        with self.assertWarnsRegex(DeprecationWarning, "Glyph.scale()"):
            glyph1.scale(-2)
        self.assertNotEqual(glyph1.bounds, glyph2.bounds)
        glyph2.scaleBy(-2)
        self.assertEqual(glyph1.bounds, glyph2.bounds)

    def test_glyph_deprecated_scale_center(self):
        glyph1 = self.getGlyph_generic()
        glyph2 = self.getGlyph_generic()
        with self.assertWarnsRegex(DeprecationWarning, "Glyph.scale()"):
            glyph1.scale((-2, 3), center=(1, 2))
        self.assertNotEqual(glyph1.bounds, glyph2.bounds)
        glyph2.scaleBy((-2, 3), origin=(1, 2))
        self.assertEqual(glyph1.bounds, glyph2.bounds)

    def test_glyph_deprecated_rotate_no_offset(self):
        glyph1 = self.getGlyph_generic()
        glyph2 = self.getGlyph_generic()
        with self.assertWarnsRegex(DeprecationWarning, "Glyph.rotate()"):
            glyph1.rotate(45)
        self.assertNotEqual(glyph1.bounds, glyph2.bounds)
        glyph2.rotateBy(45)
        self.assertEqual(glyph1.bounds, glyph2.bounds)

    def test_glyph_deprecated_rotate_offset(self):
        glyph1 = self.getGlyph_generic()
        glyph2 = self.getGlyph_generic()
        with self.assertWarnsRegex(DeprecationWarning, "Glyph.rotate()"):
            glyph1.rotate(45, offset=(1, 2))
        self.assertNotEqual(glyph1.bounds, glyph2.bounds)
        glyph2.rotateBy(45, origin=(1, 2))
        self.assertEqual(glyph1.bounds, glyph2.bounds)

    def test_glyph_deprecated_transform(self):
        glyph1 = self.getGlyph_generic()
        glyph2 = self.getGlyph_generic()
        with self.assertWarnsRegex(DeprecationWarning, "Glyph.transform()"):
            glyph1.transform((2, 0, 0, 3, -3, 2))
        self.assertNotEqual(glyph1.bounds, glyph2.bounds)
        glyph2.transformBy((2, 0, 0, 3, -3, 2))
        self.assertEqual(glyph1.bounds, glyph2.bounds)

    def test_glyph_deprecated_skew_no_offset_one_value(self):
        glyph1 = self.getGlyph_generic()
        glyph2 = self.getGlyph_generic()
        with self.assertWarnsRegex(DeprecationWarning, "Glyph.skew()"):
            glyph1.skew(100)
        self.assertNotEqual(glyph1.bounds, glyph2.bounds)
        glyph2.skewBy(100)
        self.assertEqual(glyph1.bounds, glyph2.bounds)

    def test_glyph_deprecated_skew_no_offset_two_values(self):
        glyph1 = self.getGlyph_generic()
        glyph2 = self.getGlyph_generic()
        with self.assertWarnsRegex(DeprecationWarning, "Glyph.skew()"):
            glyph1.skew((100, 200))
        self.assertNotEqual(glyph1.bounds, glyph2.bounds)
        glyph2.skewBy((100, 200))
        self.assertEqual(glyph1.bounds, glyph2.bounds)

    def test_glyph_deprecated_skew_offset_one_value(self):
        glyph1 = self.getGlyph_generic()
        glyph2 = self.getGlyph_generic()
        with self.assertWarnsRegex(DeprecationWarning, "Glyph.skew()"):
            glyph1.skew(100, offset=(1, 2))
        self.assertNotEqual(glyph1.bounds, glyph2.bounds)
        glyph2.skewBy(100, origin=(1, 2))
        self.assertEqual(glyph1.bounds, glyph2.bounds)

    def test_glyph_deprecated_skew_offset_two_values(self):
        glyph1 = self.getGlyph_generic()
        glyph2 = self.getGlyph_generic()
        with self.assertWarnsRegex(DeprecationWarning, "Glyph.skew()"):
            glyph1.skew((100, 200), offset=(1, 2))
        self.assertNotEqual(glyph1.bounds, glyph2.bounds)
        glyph2.skewBy((100, 200), origin=(1, 2))
        self.assertEqual(glyph1.bounds, glyph2.bounds)

    # -------
    # Contour
    # -------

    def getContour_bounds(self):
        contour, _ = self.objectGenerator("contour")
        contour.appendPoint((0, 0), "line")
        contour.appendPoint((0, 100), "line")
        contour.appendPoint((100, 100), "line")
        contour.appendPoint((100, 0), "line")
        return contour

    def test_contour_removed_setParent(self):
        glyph, _ = self.objectGenerator("glyph")
        contour = self.getContour_bounds()
        with self.assertRaisesRegex(RemovedError, "setParent()"):
            contour.setParent(glyph)

    def test_contour_deprecated__get_box(self):
        contour = self.getContour_bounds()
        with self.assertWarnsRegex(DeprecationWarning, "Contour.bounds"):
            box = contour._get_box()
        self.assertEqual(box, contour.bounds)

    def test_contour_deprecated_box(self):
        contour = self.getContour_bounds()
        with self.assertWarnsRegex(DeprecationWarning, "Contour.bounds"):
            box = contour.box
        self.assertEqual(box, contour.bounds)

    def test_contour_deprecated_reverseContour(self):
        contour1 = self.getContour_bounds()
        contour2 = self.getContour_bounds()
        self.assertEqual(contour1.clockwise, contour2.clockwise)
        with self.assertWarnsRegex(DeprecationWarning, "Contour.reverse()"):
            contour1.reverseContour()
        self.assertNotEqual(contour1.clockwise, contour2.clockwise)
        contour2.reverse()
        self.assertEqual(contour1.clockwise, contour2.clockwise)

    def test_contour_deprecated__generateIdentifer(self):
        contour, _ = self.objectGenerator("contour")
        with self.assertWarnsRegex(DeprecationWarning, "Contour._generateIdentifier()"):
            i = contour._generateIdentifier()
        self.assertEqual(i, contour._getIdentifier())

    def test_contour_deprecated_generateIdentifer(self):
        contour, _ = self.objectGenerator("contour")
        with self.assertWarnsRegex(DeprecationWarning, "Contour.generateIdentifier()"):
            i = contour.generateIdentifier()
        self.assertEqual(i, contour.getIdentifier())

    def test_contour_deprecated__generateIdentiferforPoint(self):
        contour = self.getContour_bounds()
        with self.assertWarnsRegex(
            DeprecationWarning, "Contour._generateIdentifierforPoint()"
        ):
            i = contour._generateIdentifierforPoint(contour[0][0])
        self.assertEqual(i, contour._getIdentifierForPoint(contour[0][0]))

    def test_contour_deprecated_generateIdentiferForPoint(self):
        contour = self.getContour_bounds()
        with self.assertWarnsRegex(
            DeprecationWarning, "Contour.generateIdentifierforPoint()"
        ):
            i = contour.generateIdentifierforPoint(contour[0][0])
        self.assertEqual(i, contour.getIdentifierForPoint(contour[0][0]))

    def test_contour_deprecated_getParent(self):
        glyph, _ = self.objectGenerator("glyph")
        contour = self.getContour_bounds()
        contour.glyph = glyph
        with self.assertWarnsRegex(DeprecationWarning, "Contour.glyph"):
            p = contour.getParent()
        self.assertEqual(p, contour.glyph)

    def test_contour_deprecated_update(self):
        # As changed() is defined by the environment, only test if a Warning is issued.
        contour, _ = self.objectGenerator("contour")
        with self.assertWarnsRegex(DeprecationWarning, "Contour.changed()"):
            contour.update()

    def test_contour_deprecated_setChanged(self):
        # As changed() is defined by the environment, only test if a Warning is issued.
        contour, _ = self.objectGenerator("contour")
        with self.assertWarnsRegex(DeprecationWarning, "Contour.changed()"):
            contour.setChanged()

    def test_contour_deprecated_move(self):
        contour1 = self.getContour_bounds()
        contour2 = self.getContour_bounds()
        with self.assertWarnsRegex(DeprecationWarning, "Contour.move()"):
            contour1.move((0, 20))
        self.assertNotEqual(contour1.bounds, contour2.bounds)
        contour2.moveBy((0, 20))
        self.assertEqual(contour1.bounds, contour2.bounds)

    def test_contour_deprecated_translate(self):
        contour1 = self.getContour_bounds()
        contour2 = self.getContour_bounds()
        with self.assertWarnsRegex(DeprecationWarning, "Contour.translate()"):
            contour1.translate((0, 20))
        self.assertNotEqual(contour1.bounds, contour2.bounds)
        contour2.moveBy((0, 20))
        self.assertEqual(contour1.bounds, contour2.bounds)

    def test_contour_deprecated_scale_no_center(self):
        contour1 = self.getContour_bounds()
        contour2 = self.getContour_bounds()
        with self.assertWarnsRegex(DeprecationWarning, "Contour.scale()"):
            contour1.scale(-2)
        self.assertNotEqual(contour1.bounds, contour2.bounds)
        contour2.scaleBy(-2)
        self.assertEqual(contour1.bounds, contour2.bounds)

    def test_contour_deprecated_scale_center(self):
        contour1 = self.getContour_bounds()
        contour2 = self.getContour_bounds()
        with self.assertWarnsRegex(DeprecationWarning, "Contour.scale()"):
            contour1.scale((-2, 3), center=(1, 2))
        self.assertNotEqual(contour1.bounds, contour2.bounds)
        contour2.scaleBy((-2, 3), origin=(1, 2))
        self.assertEqual(contour1.bounds, contour2.bounds)

    def test_contour_deprecated_rotate_no_offset(self):
        contour1 = self.getContour_bounds()
        contour2 = self.getContour_bounds()
        with self.assertWarnsRegex(DeprecationWarning, "Contour.rotate()"):
            contour1.rotate(45)
        self.assertNotEqual(contour1.bounds, contour2.bounds)
        contour2.rotateBy(45)
        self.assertEqual(contour1.bounds, contour2.bounds)

    def test_contour_deprecated_rotate_offset(self):
        contour1 = self.getContour_bounds()
        contour2 = self.getContour_bounds()
        with self.assertWarnsRegex(DeprecationWarning, "Contour.rotate()"):
            contour1.rotate(45, offset=(1, 2))
        self.assertNotEqual(contour1.bounds, contour2.bounds)
        contour2.rotateBy(45, origin=(1, 2))
        self.assertEqual(contour1.bounds, contour2.bounds)

    def test_contour_deprecated_transform(self):
        contour1 = self.getContour_bounds()
        contour2 = self.getContour_bounds()
        with self.assertWarnsRegex(DeprecationWarning, "Contour.transform()"):
            contour1.transform((2, 0, 0, 3, -3, 2))
        self.assertNotEqual(contour1.bounds, contour2.bounds)
        contour2.transformBy((2, 0, 0, 3, -3, 2))
        self.assertEqual(contour1.bounds, contour2.bounds)

    def test_contour_deprecated_skew_no_offset_one_value(self):
        contour1 = self.getContour_bounds()
        contour2 = self.getContour_bounds()
        with self.assertWarnsRegex(DeprecationWarning, "Contour.skew()"):
            contour1.skew(100)
        self.assertNotEqual(contour1.bounds, contour2.bounds)
        contour2.skewBy(100)
        self.assertEqual(contour1.bounds, contour2.bounds)

    def test_contour_deprecated_skew_no_offset_two_values(self):
        contour1 = self.getContour_bounds()
        contour2 = self.getContour_bounds()
        with self.assertWarnsRegex(DeprecationWarning, "Contour.skew()"):
            contour1.skew((100, 200))
        self.assertNotEqual(contour1.bounds, contour2.bounds)
        contour2.skewBy((100, 200))
        self.assertEqual(contour1.bounds, contour2.bounds)

    def test_contour_deprecated_skew_offset_one_value(self):
        contour1 = self.getContour_bounds()
        contour2 = self.getContour_bounds()
        with self.assertWarnsRegex(DeprecationWarning, "Contour.skew()"):
            contour1.skew(100, offset=(1, 2))
        self.assertNotEqual(contour1.bounds, contour2.bounds)
        contour2.skewBy(100, origin=(1, 2))
        self.assertEqual(contour1.bounds, contour2.bounds)

    def test_contour_deprecated_skew_offset_two_values(self):
        contour1 = self.getContour_bounds()
        contour2 = self.getContour_bounds()
        with self.assertWarnsRegex(DeprecationWarning, "Contour.skew()"):
            contour1.skew((100, 200), offset=(1, 2))
        self.assertNotEqual(contour1.bounds, contour2.bounds)
        contour2.skewBy((100, 200), origin=(1, 2))
        self.assertEqual(contour1.bounds, contour2.bounds)

    # -------
    # Segment
    # -------

    def getSegment(self):
        contour, unrequested = self.objectGenerator("contour")
        unrequested.append(contour)
        contour.appendPoint((0, 0), "move")
        contour.appendPoint((101, 202), "line")
        segment = contour[1]
        return segment

    def test_segment_removed_insertPoint(self):
        segment = self.getSegment()
        with self.assertRaisesRegex(RemovedError, "insertPoint()"):
            segment.insertPoint(None)

    def test_segment_removed_removePoint(self):
        segment = self.getSegment()
        with self.assertRaisesRegex(RemovedError, "removePoint()"):
            segment.removePoint(None)

    def test_segment_removed_setParent(self):
        segment = self.getSegment()
        with self.assertRaisesRegex(RemovedError, "setParent()"):
            segment.setParent(None)

    def test_segment_deprecated_getParent(self):
        segment = self.getSegment()
        with self.assertWarnsRegex(DeprecationWarning, "Segment.contour"):
            c = segment.getParent()
        self.assertEqual(c, segment.contour)

    def test_segment_deprecated_update(self):
        # As changed() is defined by the environment, only test if a Warning is issued.
        segment = self.getSegment()
        with self.assertWarnsRegex(DeprecationWarning, "Segment.changed()"):
            segment.update()

    def test_segment_deprecated_setChanged(self):
        # As changed() is defined by the environment, only test if a Warning is issued.
        segment = self.getSegment()
        with self.assertWarnsRegex(DeprecationWarning, "Segment.changed()"):
            segment.setChanged()

    def test_segment_deprecated_move(self):
        segment1 = self.getSegment()
        segment2 = self.getSegment()
        with self.assertWarnsRegex(DeprecationWarning, "Segment.move()"):
            segment1.move((0, 20))
        coordinates1 = tuple((point.x, point.y) for point in segment1.points)
        coordinates2 = tuple((point.x, point.y) for point in segment2.points)
        self.assertNotEqual(coordinates1, coordinates2)
        segment2.moveBy((0, 20))
        coordinates2 = tuple((point.x, point.y) for point in segment2.points)
        self.assertEqual(coordinates1, coordinates2)

    def test_segment_deprecated_translate(self):
        segment1 = self.getSegment()
        segment2 = self.getSegment()
        with self.assertWarnsRegex(DeprecationWarning, "Segment.translate()"):
            segment1.translate((0, 20))
        coordinates1 = tuple((point.x, point.y) for point in segment1.points)
        coordinates2 = tuple((point.x, point.y) for point in segment2.points)
        self.assertNotEqual(coordinates1, coordinates2)
        segment2.moveBy((0, 20))
        coordinates2 = tuple((point.x, point.y) for point in segment2.points)
        self.assertEqual(coordinates1, coordinates2)

    def test_segment_deprecated_scale_no_center(self):
        segment1 = self.getSegment()
        segment2 = self.getSegment()
        with self.assertWarnsRegex(DeprecationWarning, "Segment.scale()"):
            segment1.scale(-2)
        coordinates1 = tuple((point.x, point.y) for point in segment1.points)
        coordinates2 = tuple((point.x, point.y) for point in segment2.points)
        self.assertNotEqual(coordinates1, coordinates2)
        segment2.scaleBy(-2)
        coordinates2 = tuple((point.x, point.y) for point in segment2.points)
        self.assertEqual(coordinates1, coordinates2)

    def test_segment_deprecated_scale_center(self):
        segment1 = self.getSegment()
        segment2 = self.getSegment()
        with self.assertWarnsRegex(DeprecationWarning, "Segment.scale()"):
            segment1.scale((-2, 3), center=(1, 2))
        coordinates1 = tuple((point.x, point.y) for point in segment1.points)
        coordinates2 = tuple((point.x, point.y) for point in segment2.points)
        self.assertNotEqual(coordinates1, coordinates2)
        segment2.scaleBy((-2, 3), origin=(1, 2))
        coordinates2 = tuple((point.x, point.y) for point in segment2.points)
        self.assertEqual(coordinates1, coordinates2)

    def test_segment_deprecated_rotate_no_offset(self):
        segment1 = self.getSegment()
        segment2 = self.getSegment()
        with self.assertWarnsRegex(DeprecationWarning, "Segment.rotate()"):
            segment1.rotate(45)
        coordinates1 = tuple((point.x, point.y) for point in segment1.points)
        coordinates2 = tuple((point.x, point.y) for point in segment2.points)
        self.assertNotEqual(coordinates1, coordinates2)
        segment2.rotateBy(45)
        coordinates2 = tuple((point.x, point.y) for point in segment2.points)
        self.assertEqual(coordinates1, coordinates2)

    def test_segment_deprecated_rotate_offset(self):
        segment1 = self.getSegment()
        segment2 = self.getSegment()
        with self.assertWarnsRegex(DeprecationWarning, "Segment.rotate()"):
            segment1.rotate(45, offset=(1, 2))
        coordinates1 = tuple((point.x, point.y) for point in segment1.points)
        coordinates2 = tuple((point.x, point.y) for point in segment2.points)
        self.assertNotEqual(coordinates1, coordinates2)
        segment2.rotateBy(45, origin=(1, 2))
        coordinates2 = tuple((point.x, point.y) for point in segment2.points)
        self.assertEqual(coordinates1, coordinates2)

    def test_segment_deprecated_transform(self):
        segment1 = self.getSegment()
        segment2 = self.getSegment()
        with self.assertWarnsRegex(DeprecationWarning, "Segment.transform()"):
            segment1.transform((2, 0, 0, 3, -3, 2))
        coordinates1 = tuple((point.x, point.y) for point in segment1.points)
        coordinates2 = tuple((point.x, point.y) for point in segment2.points)
        self.assertNotEqual(coordinates1, coordinates2)
        segment2.transformBy((2, 0, 0, 3, -3, 2))
        coordinates2 = tuple((point.x, point.y) for point in segment2.points)
        self.assertEqual(coordinates1, coordinates2)

    def test_segment_deprecated_skew_no_offset_one_value(self):
        segment1 = self.getSegment()
        segment2 = self.getSegment()
        with self.assertWarnsRegex(DeprecationWarning, "Segment.skew()"):
            segment1.skew(100)
        coordinates1 = tuple((point.x, point.y) for point in segment1.points)
        coordinates2 = tuple((point.x, point.y) for point in segment2.points)
        self.assertNotEqual(coordinates1, coordinates2)
        segment2.skewBy(100)
        coordinates2 = tuple((point.x, point.y) for point in segment2.points)
        self.assertEqual(coordinates1, coordinates2)

    def test_segment_deprecated_skew_no_offset_two_values(self):
        segment1 = self.getSegment()
        segment2 = self.getSegment()
        with self.assertWarnsRegex(DeprecationWarning, "Segment.skew()"):
            segment1.skew((100, 200))
        coordinates1 = tuple((point.x, point.y) for point in segment1.points)
        coordinates2 = tuple((point.x, point.y) for point in segment2.points)
        self.assertNotEqual(coordinates1, coordinates2)
        segment2.skewBy((100, 200))
        coordinates2 = tuple((point.x, point.y) for point in segment2.points)
        self.assertEqual(coordinates1, coordinates2)

    def test_segment_deprecated_skew_offset_one_value(self):
        segment1 = self.getSegment()
        segment2 = self.getSegment()
        with self.assertWarnsRegex(DeprecationWarning, "Segment.skew()"):
            segment1.skew(100, offset=(1, 2))
        coordinates1 = tuple((point.x, point.y) for point in segment1.points)
        coordinates2 = tuple((point.x, point.y) for point in segment2.points)
        self.assertNotEqual(coordinates1, coordinates2)
        segment2.skewBy(100, origin=(1, 2))
        coordinates2 = tuple((point.x, point.y) for point in segment2.points)
        self.assertEqual(coordinates1, coordinates2)

    def test_segment_deprecated_skew_offset_two_values(self):
        segment1 = self.getSegment()
        segment2 = self.getSegment()
        with self.assertWarnsRegex(DeprecationWarning, "Segment.skew()"):
            segment1.skew((100, 200), offset=(1, 2))
        coordinates1 = tuple((point.x, point.y) for point in segment1.points)
        coordinates2 = tuple((point.x, point.y) for point in segment2.points)
        self.assertNotEqual(coordinates1, coordinates2)
        segment2.skewBy((100, 200), origin=(1, 2))
        coordinates2 = tuple((point.x, point.y) for point in segment2.points)
        self.assertEqual(coordinates1, coordinates2)

    # ---------
    # Component
    # ---------

    def getComponent(self):
        layer, _ = self.objectGenerator("layer")
        glyph = layer.newGlyph("A")
        pen = glyph.getPen()
        pen.moveTo((0, 0))
        pen.lineTo((0, 100))
        pen.lineTo((100, 100))
        pen.lineTo((100, 0))
        pen.closePath()
        for i, point in enumerate(glyph[0].points):
            point.name = f"point {i}"
        glyph = layer.newGlyph("B")
        component = glyph.appendComponent("A")
        component.transformation = (1, 2, 3, 4, 5, 6)
        return component

    def test_component_removed_setParent(self):
        component = self.getComponent()
        with self.assertRaisesRegex(RemovedError, "setParent()"):
            component.setParent(None)

    def test_component_deprecated__get_box(self):
        component = self.getComponent()
        with self.assertWarnsRegex(DeprecationWarning, "Component.bounds"):
            box = component._get_box()
        self.assertEqual(box, component.bounds)

    def test_component_deprecated_box(self):
        component = self.getComponent()
        with self.assertWarnsRegex(DeprecationWarning, "Component.bounds"):
            box = component.box
        self.assertEqual(box, component.bounds)

    def test_component_deprecated__generateIdentifier(self):
        component = self.getComponent()
        with self.assertWarnsRegex(DeprecationWarning, "Component._getIdentifier()"):
            i = component._generateIdentifier()
        self.assertEqual(i, component._getIdentifier())

    def test_component_deprecated_generateIdentifier(self):
        component = self.getComponent()
        with self.assertWarnsRegex(DeprecationWarning, "Component.getIdentifier()"):
            i = component.generateIdentifier()
        self.assertEqual(i, component.getIdentifier())

    def test_component_deprecated_getParent(self):
        component = self.getComponent()
        with self.assertWarnsRegex(DeprecationWarning, "Component.glyph"):
            p = component.getParent()
        self.assertEqual(p, component.glyph)

    def test_component_deprecated_update(self):
        # As changed() is defined by the environment, only test if a Warning is issued.
        component = self.getComponent()
        with self.assertWarnsRegex(DeprecationWarning, "Component.changed()"):
            component.update()

    def test_component_deprecated_setChanged(self):
        # As changed() is defined by the environment, only test if a Warning is issued.
        component = self.getComponent()
        with self.assertWarnsRegex(DeprecationWarning, "Component.changed()"):
            component.setChanged()

    def test_component_deprecated_move(self):
        component1 = self.getComponent()
        component2 = self.getComponent()
        with self.assertWarnsRegex(DeprecationWarning, "Component.move()"):
            component1.move((0, 20))
        self.assertNotEqual(component1.bounds, component2.bounds)
        component2.moveBy((0, 20))
        self.assertEqual(component1.bounds, component2.bounds)

    def test_component_deprecated_translate(self):
        component1 = self.getComponent()
        component2 = self.getComponent()
        with self.assertWarnsRegex(DeprecationWarning, "Component.translate()"):
            component1.translate((0, 20))
        self.assertNotEqual(component1.bounds, component2.bounds)
        component2.moveBy((0, 20))
        self.assertEqual(component1.bounds, component2.bounds)

    def test_component_deprecated_rotate_no_offset(self):
        component1 = self.getComponent()
        component2 = self.getComponent()
        with self.assertWarnsRegex(DeprecationWarning, "Component.rotate()"):
            component1.rotate(45)
        self.assertNotEqual(component1.bounds, component2.bounds)
        component2.rotateBy(45)
        self.assertEqual(component1.bounds, component2.bounds)

    def test_component_deprecated_rotate_offset(self):
        component1 = self.getComponent()
        component2 = self.getComponent()
        with self.assertWarnsRegex(DeprecationWarning, "Component.rotate()"):
            component1.rotate(45, offset=(1, 2))
        self.assertNotEqual(component1.bounds, component2.bounds)
        component2.rotateBy(45, origin=(1, 2))
        self.assertEqual(component1.bounds, component2.bounds)

    def test_component_deprecated_transform(self):
        component1 = self.getComponent()
        component2 = self.getComponent()
        with self.assertWarnsRegex(DeprecationWarning, "Component.transform()"):
            component1.transform((2, 0, 0, 3, -3, 2))
        self.assertNotEqual(component1.bounds, component2.bounds)
        component2.transformBy((2, 0, 0, 3, -3, 2))
        self.assertEqual(component1.bounds, component2.bounds)

    def test_component_deprecated_skew_no_offset_one_value(self):
        component1 = self.getComponent()
        component2 = self.getComponent()
        with self.assertWarnsRegex(DeprecationWarning, "Component.skew()"):
            component1.skew(100)
        self.assertNotEqual(component1.bounds, component2.bounds)
        component2.skewBy(100)
        self.assertEqual(component1.bounds, component2.bounds)

    def test_component_deprecated_skew_no_offset_two_values(self):
        component1 = self.getComponent()
        component2 = self.getComponent()
        with self.assertWarnsRegex(DeprecationWarning, "Component.skew()"):
            component1.skew((100, 200))
        self.assertNotEqual(component1.bounds, component2.bounds)
        component2.skewBy((100, 200))
        self.assertEqual(component1.bounds, component2.bounds)

    def test_component_deprecated_skew_offset_one_value(self):
        component1 = self.getComponent()
        component2 = self.getComponent()
        with self.assertWarnsRegex(DeprecationWarning, "Component.skew()"):
            component1.skew(100, offset=(1, 2))
        self.assertNotEqual(component1.bounds, component2.bounds)
        component2.skewBy(100, origin=(1, 2))
        self.assertEqual(component1.bounds, component2.bounds)

    def test_component_deprecated_skew_offset_two_values(self):
        component1 = self.getComponent()
        component2 = self.getComponent()
        with self.assertWarnsRegex(DeprecationWarning, "Component.skew()"):
            component1.skew((100, 200), offset=(1, 2))
        self.assertNotEqual(component1.bounds, component2.bounds)
        component2.skewBy((100, 200), origin=(1, 2))
        self.assertEqual(component1.bounds, component2.bounds)

    # -----
    # Point
    # -----

    def getPoint(self):
        contour, _ = self.objectGenerator("contour")
        contour.appendPoint((0, 0), "move")
        contour.appendPoint((101, 202), "line")
        point = contour.points[1]
        point.smooth = True
        return point

    def test_point_removed_select(self):
        point = self.getPoint()
        with self.assertRaisesRegex(RemovedError, "Point.select"):
            point.select()

    def test_point_removed_setParent(self):
        point = self.getPoint()
        with self.assertRaisesRegex(RemovedError, "setParent()"):
            point.setParent(None)

    def test_point_deprecated__generateIdentifier(self):
        point = self.getPoint()
        with self.assertWarnsRegex(DeprecationWarning, "Point._getIdentifier()"):
            i = point._generateIdentifier()
        self.assertEqual(i, point._getIdentifier())

    def test_point_deprecated_generateIdentifier(self):
        point = self.getPoint()
        with self.assertWarnsRegex(DeprecationWarning, "Point.getIdentifier()"):
            i = point.generateIdentifier()
        self.assertEqual(i, point.getIdentifier())

    def test_point_getParent(self):
        point = self.getPoint()
        with self.assertWarnsRegex(DeprecationWarning, "Point.contour"):
            p = point.getParent()
        self.assertEqual(p, point.contour)

    def test_point_deprecated_update(self):
        # As changed() is defined by the environment, only test if a Warning is issued.
        point = self.getPoint()
        with self.assertWarnsRegex(DeprecationWarning, "Point.changed()"):
            point.update()

    def test_point_deprecated_setChanged(self):
        # As changed() is defined by the environment, only test if a Warning is issued.
        point = self.getPoint()
        with self.assertWarnsRegex(DeprecationWarning, "Point.changed()"):
            point.setChanged()

    def test_point_deprecated_move(self):
        point1 = self.getPoint()
        point2 = self.getPoint()
        with self.assertWarnsRegex(DeprecationWarning, "Point.move()"):
            point1.move((0, 20))
        self.assertNotEqual((point1.x, point1.y), (point2.x, point2.y))
        point2.moveBy((0, 20))
        self.assertEqual((point1.x, point1.y), (point2.x, point2.y))

    def test_point_deprecated_translate(self):
        point1 = self.getPoint()
        point2 = self.getPoint()
        with self.assertWarnsRegex(DeprecationWarning, "Point.translate()"):
            point1.translate((0, 20))
        self.assertNotEqual((point1.x, point1.y), (point2.x, point2.y))
        point2.moveBy((0, 20))
        self.assertEqual((point1.x, point1.y), (point2.x, point2.y))

    def test_point_deprecated_scale_no_center(self):
        point1 = self.getPoint()
        point2 = self.getPoint()
        with self.assertWarnsRegex(DeprecationWarning, "Point.scale()"):
            point1.scale(-2)
        self.assertNotEqual((point1.x, point1.y), (point2.x, point2.y))
        point2.scaleBy(-2)
        self.assertEqual((point1.x, point1.y), (point2.x, point2.y))

    def test_point_deprecated_scale_center(self):
        point1 = self.getPoint()
        point2 = self.getPoint()
        with self.assertWarnsRegex(DeprecationWarning, "Point.scale()"):
            point1.scale((-2, 3), center=(1, 2))
        self.assertNotEqual((point1.x, point1.y), (point2.x, point2.y))
        point2.scaleBy((-2, 3), origin=(1, 2))
        self.assertEqual((point1.x, point1.y), (point2.x, point2.y))

    def test_point_deprecated_rotate_no_offset(self):
        point1 = self.getPoint()
        point2 = self.getPoint()
        with self.assertWarnsRegex(DeprecationWarning, "Point.rotate()"):
            point1.rotate(45)
        self.assertNotEqual((point1.x, point1.y), (point2.x, point2.y))
        point2.rotateBy(45)
        self.assertEqual((point1.x, point1.y), (point2.x, point2.y))

    def test_point_deprecated_rotate_offset(self):
        point1 = self.getPoint()
        point2 = self.getPoint()
        with self.assertWarnsRegex(DeprecationWarning, "Point.rotate()"):
            point1.rotate(45, offset=(1, 2))
        self.assertNotEqual((point1.x, point1.y), (point2.x, point2.y))
        point2.rotateBy(45, origin=(1, 2))
        self.assertEqual((point1.x, point1.y), (point2.x, point2.y))

    def test_point_deprecated_transform(self):
        point1 = self.getPoint()
        point2 = self.getPoint()
        with self.assertWarnsRegex(DeprecationWarning, "Point.transform()"):
            point1.transform((2, 0, 0, 3, -3, 2))
        self.assertNotEqual((point1.x, point1.y), (point2.x, point2.y))
        point2.transformBy((2, 0, 0, 3, -3, 2))
        self.assertEqual((point1.x, point1.y), (point2.x, point2.y))

    def test_point_deprecated_skew_no_offset_one_value(self):
        point1 = self.getPoint()
        point2 = self.getPoint()
        with self.assertWarnsRegex(DeprecationWarning, "Point.skew()"):
            point1.skew(100)
        self.assertNotEqual((point1.x, point1.y), (point2.x, point2.y))
        point2.skewBy(100)
        self.assertEqual((point1.x, point1.y), (point2.x, point2.y))

    def test_point_deprecated_skew_no_offset_two_values(self):
        point1 = self.getPoint()
        point2 = self.getPoint()
        with self.assertWarnsRegex(DeprecationWarning, "Point.skew()"):
            point1.skew((100, 200))
        self.assertNotEqual((point1.x, point1.y), (point2.x, point2.y))
        point2.skewBy((100, 200))
        self.assertEqual((point1.x, point1.y), (point2.x, point2.y))

    def test_point_deprecated_skew_offset_one_value(self):
        point1 = self.getPoint()
        point2 = self.getPoint()
        with self.assertWarnsRegex(DeprecationWarning, "Point.skew()"):
            point1.skew(100, offset=(1, 2))
        self.assertNotEqual((point1.x, point1.y), (point2.x, point2.y))
        point2.skewBy(100, origin=(1, 2))
        self.assertEqual((point1.x, point1.y), (point2.x, point2.y))

    def test_point_deprecated_skew_offset_two_values(self):
        point1 = self.getPoint()
        point2 = self.getPoint()
        with self.assertWarnsRegex(DeprecationWarning, "Point.skew()"):
            point1.skew((100, 200), offset=(1, 2))
        self.assertNotEqual((point1.x, point1.y), (point2.x, point2.y))
        point2.skewBy((100, 200), origin=(1, 2))
        self.assertEqual((point1.x, point1.y), (point2.x, point2.y))

    # ------
    # bPoint
    # ------

    def getBPoint(self):
        contour, _ = self.objectGenerator("contour")
        contour.appendPoint((0, 0), "move")
        contour.appendPoint((101, 202), "line")
        contour.appendPoint((303, 0), "line")
        bPoint = contour.bPoints[1]
        return bPoint

    def test_bPoint_removed_select(self):
        bPoint = self.getBPoint()
        with self.assertRaisesRegex(RemovedError, "BPoint.select"):
            bPoint.select()

    def test_bPoint_removed_setParent(self):
        bPoint = self.getBPoint()
        with self.assertRaisesRegex(RemovedError, "setParent()"):
            bPoint.setParent(None)

    def test_bPoint_deprecated__generateIdentifier(self):
        bPoint = self.getBPoint()
        with self.assertWarnsRegex(DeprecationWarning, "BPoint._getIdentifier()"):
            i = bPoint._generateIdentifier()
        self.assertEqual(i, bPoint._getIdentifier())

    def test_bPoint_deprecated_generateIdentifier(self):
        bPoint = self.getBPoint()
        with self.assertWarnsRegex(DeprecationWarning, "BPoint.getIdentifier()"):
            i = bPoint.generateIdentifier()
        self.assertEqual(i, bPoint.getIdentifier())

    def test_bPoint_getParent(self):
        bPoint = self.getBPoint()
        with self.assertWarnsRegex(DeprecationWarning, "BPoint.contour"):
            p = bPoint.getParent()
        self.assertEqual(p, bPoint.contour)

    def test_bPoint_deprecated_update(self):
        # As changed() is defined by the environment, only test if a Warning is issued.
        bPoint = self.getBPoint()
        with self.assertWarnsRegex(DeprecationWarning, "BPoint.changed()"):
            bPoint.update()

    def test_bPoint_deprecated_setChanged(self):
        # As changed() is defined by the environment, only test if a Warning is issued.
        bPoint = self.getBPoint()
        with self.assertWarnsRegex(DeprecationWarning, "BPoint.changed()"):
            bPoint.setChanged()

    def test_bPoint_deprecated_move(self):
        bPoint1 = self.getBPoint()
        bPoint2 = self.getBPoint()
        with self.assertWarnsRegex(DeprecationWarning, "BPoint.move()"):
            bPoint1.move((0, 20))
        self.assertNotEqual(bPoint1.anchor, bPoint2.anchor)
        bPoint2.moveBy((0, 20))
        self.assertEqual(bPoint1.anchor, bPoint2.anchor)

    def test_bPoint_deprecated_translate(self):
        bPoint1 = self.getBPoint()
        bPoint2 = self.getBPoint()
        with self.assertWarnsRegex(DeprecationWarning, "BPoint.translate()"):
            bPoint1.translate((0, 20))
        self.assertNotEqual(bPoint1.anchor, bPoint2.anchor)
        bPoint2.moveBy((0, 20))
        self.assertEqual(bPoint1.anchor, bPoint2.anchor)

    def test_bPoint_deprecated_scale_no_center(self):
        bPoint1 = self.getBPoint()
        bPoint2 = self.getBPoint()
        with self.assertWarnsRegex(DeprecationWarning, "BPoint.scale()"):
            bPoint1.scale(-2)
        self.assertNotEqual(bPoint1.anchor, bPoint2.anchor)
        bPoint2.scaleBy(-2)
        self.assertEqual(bPoint1.anchor, bPoint2.anchor)

    def test_bPoint_deprecated_scale_center(self):
        bPoint1 = self.getBPoint()
        bPoint2 = self.getBPoint()
        with self.assertWarnsRegex(DeprecationWarning, "BPoint.scale()"):
            bPoint1.scale((-2, 3), center=(1, 2))
        self.assertNotEqual(bPoint1.anchor, bPoint2.anchor)
        bPoint2.scaleBy((-2, 3), origin=(1, 2))
        self.assertEqual(bPoint1.anchor, bPoint2.anchor)

    def test_bPoint_deprecated_rotate_no_offset(self):
        bPoint1 = self.getBPoint()
        bPoint2 = self.getBPoint()
        with self.assertWarnsRegex(DeprecationWarning, "BPoint.rotate()"):
            bPoint1.rotate(45)
        self.assertNotEqual(bPoint1.anchor, bPoint2.anchor)
        bPoint2.rotateBy(45)
        self.assertEqual(bPoint1.anchor, bPoint2.anchor)

    def test_bPoint_deprecated_rotate_offset(self):
        bPoint1 = self.getBPoint()
        bPoint2 = self.getBPoint()
        with self.assertWarnsRegex(DeprecationWarning, "BPoint.rotate()"):
            bPoint1.rotate(45, offset=(1, 2))
        self.assertNotEqual(bPoint1.anchor, bPoint2.anchor)
        bPoint2.rotateBy(45, origin=(1, 2))
        self.assertEqual(bPoint1.anchor, bPoint2.anchor)

    def test_bPoint_deprecated_transform(self):
        bPoint1 = self.getBPoint()
        bPoint2 = self.getBPoint()
        with self.assertWarnsRegex(DeprecationWarning, "BPoint.transform()"):
            bPoint1.transform((2, 0, 0, 3, -3, 2))
        self.assertNotEqual(bPoint1.anchor, bPoint2.anchor)
        bPoint2.transformBy((2, 0, 0, 3, -3, 2))
        self.assertEqual(bPoint1.anchor, bPoint2.anchor)

    def test_bPoint_deprecated_skew_no_offset_one_value(self):
        bPoint1 = self.getBPoint()
        bPoint2 = self.getBPoint()
        with self.assertWarnsRegex(DeprecationWarning, "BPoint.skew()"):
            bPoint1.skew(100)
        self.assertNotEqual(bPoint1.anchor, bPoint2.anchor)
        bPoint2.skewBy(100)
        self.assertEqual(bPoint1.anchor, bPoint2.anchor)

    def test_bPoint_deprecated_skew_no_offset_two_values(self):
        bPoint1 = self.getBPoint()
        bPoint2 = self.getBPoint()
        with self.assertWarnsRegex(DeprecationWarning, "BPoint.skew()"):
            bPoint1.skew((100, 200))
        self.assertNotEqual(bPoint1.anchor, bPoint2.anchor)
        bPoint2.skewBy((100, 200))
        self.assertEqual(bPoint1.anchor, bPoint2.anchor)

    def test_bPoint_deprecated_skew_offset_one_value(self):
        bPoint1 = self.getBPoint()
        bPoint2 = self.getBPoint()
        with self.assertWarnsRegex(DeprecationWarning, "BPoint.skew()"):
            bPoint1.skew(100, offset=(1, 2))
        self.assertNotEqual(bPoint1.anchor, bPoint2.anchor)
        bPoint2.skewBy(100, origin=(1, 2))
        self.assertEqual(bPoint1.anchor, bPoint2.anchor)

    def test_bPoint_deprecated_skew_offset_two_values(self):
        bPoint1 = self.getBPoint()
        bPoint2 = self.getBPoint()
        with self.assertWarnsRegex(DeprecationWarning, "BPoint.skew()"):
            bPoint1.skew((100, 200), offset=(1, 2))
        self.assertNotEqual(bPoint1.anchor, bPoint2.anchor)
        bPoint2.skewBy((100, 200), origin=(1, 2))
        self.assertEqual(bPoint1.anchor, bPoint2.anchor)



try:
    from ._version import __version__
except ImportError:
    try:
        from setuptools_scm import get_version

        __version__ = get_version()
    except ImportError:
        __version__ = "unknown"



import unittest
import collections


class TestFeatures(unittest.TestCase):
    def getFeatures_generic(self):
        features, _ = self.objectGenerator("features")
        features.text = "# test"
        return features

    # ----
    # repr
    # ----

    def test_reprContents(self):
        font, _ = self.objectGenerator("font")
        features = font.features
        features.text = "# test"
        value = features._reprContents()
        self.assertIsInstance(value, list)
        for i in value:
            self.assertIsInstance(i, str)

    def test_reprContents_noFont(self):
        features, _ = self.objectGenerator("features")
        features.text = "# test"
        value = features._reprContents()
        self.assertIsInstance(value, list)
        self.assertListEqual(value, [])

    # ----
    # Text
    # ----

    def test_text_get(self):
        features = self.getFeatures_generic()
        self.assertEqual(features.text, "# test")

    def test_text_valid_set(self):
        features = self.getFeatures_generic()
        features.text = "# foo"
        self.assertEqual(features.text, "# foo")

    def test_text_set_none(self):
        features = self.getFeatures_generic()
        features.text = None
        self.assertIsNone(features.text)

    def test_text_invalid_set(self):
        features = self.getFeatures_generic()
        with self.assertRaises(TypeError):
            features.text = 123

    # -------
    # Parents
    # -------

    def test_get_parent_font(self):
        font, _ = self.objectGenerator("font")
        features = font.features
        features.text = "# Test"
        self.assertIsNotNone(features.font)
        self.assertEqual(features.font, font)

    def test_get_parent_noFont(self):
        features = self.getFeatures_generic()
        self.assertIsNone(features.font)

    def test_set_parent_font(self):
        font, _ = self.objectGenerator("font")
        features = self.getFeatures_generic()
        features.font = font
        self.assertIsNotNone(features.font)
        self.assertEqual(features.font, font)

    def test_set_parent_font_none(self):
        features = self.getFeatures_generic()
        features.font = None
        self.assertIsNone(features.font)

    def test_set_parent_font_exists(self):
        font, _ = self.objectGenerator("font")
        otherFont, _ = self.objectGenerator("font")
        features = font.features
        features.text = "# Test"
        with self.assertRaises(AssertionError):
            features.font = otherFont

    # ----
    # Hash
    # ----

    def test_hash(self):
        features = self.getFeatures_generic()
        self.assertEqual(isinstance(features, collections.abc.Hashable), True)

    # --------
    # Equality
    # --------

    def test_object_equal_self(self):
        features_one = self.getFeatures_generic()
        self.assertEqual(features_one, features_one)

    def test_object_not_equal_other(self):
        features_one = self.getFeatures_generic()
        features_two = self.getFeatures_generic()
        self.assertNotEqual(features_one, features_two)

    def test_object_equal_self_variable_assignment(self):
        features_one = self.getFeatures_generic()
        a = features_one
        a.text += "# testing"
        self.assertEqual(features_one, a)

    def test_object_not_equal_self_variable_assignment(self):
        features_one = self.getFeatures_generic()
        features_two = self.getFeatures_generic()
        a = features_one
        self.assertNotEqual(features_two, a)



from __future__ import annotations
from typing import TYPE_CHECKING, Any
from collections.abc import Sequence

from fontParts.base.base import dynamicProperty
from fontParts.base.annotations import DiffType

if TYPE_CHECKING:
    from fontParts.base.base import BaseObject
    from fontParts.base.font import BaseFont
    from fontParts.base.layer import BaseLayer
    from fontParts.base.glyph import BaseGlyph
    from fontParts.base.contour import BaseContour
    from fontParts.base.component import BaseComponent
    from fontParts.base.anchor import BaseAnchor
    from fontParts.base.guideline import BaseGuideline

# ----
# Base
# ----


class BaseCompatibilityReporter:
    objectName = "Base"

    def __init__(self, obj1: BaseObject, obj2: BaseObject) -> None:
        self._object1 = obj1
        self._object2 = obj2

    # status

    fatal = False
    warning = False

    def _get_title(self) -> str:
        title = f"{self.object1Name} + {self.object2Name}"
        if self.fatal:
            return self.formatFatalString(title)
        elif self.warning:
            return self.formatWarningString(title)
        else:
            return self.formatOKString(title)

    title: dynamicProperty = dynamicProperty("title")

    # objects

    object1: dynamicProperty = dynamicProperty("object1")
    object1Name: dynamicProperty = dynamicProperty("object1Name")

    def _get_object1(self) -> Any:
        return self._object1

    def _get_object1Name(self) -> str:
        return self._getObjectName(self._object1)

    object2: dynamicProperty = dynamicProperty("object2")
    object2Name: dynamicProperty = dynamicProperty("object2Name")

    def _get_object2(self) -> Any:
        return self._object2

    def _get_object2Name(self) -> str:
        return self._getObjectName(self._object2)

    @staticmethod
    def _getObjectName(obj) -> str:
        if hasattr(obj, "name") and obj.name is not None:
            return f'"{obj.name}"'
        elif hasattr(obj, "identifier") and obj.identifier is not None:
            return f'"{obj.identifier}"'
        elif hasattr(obj, "index"):
            return f"[{obj.index}]"
        else:
            return f"<{id(obj)}>"

    # Report

    def __repr__(self) -> str:
        return self.report()

    def report(self, showOK: bool = False, showWarnings: bool = False) -> str:
        raise NotImplementedError

    def formatFatalString(self, text: str) -> str:
        return f"[Fatal] {self.objectName}: " + text

    def formatWarningString(self, text: str) -> str:
        return f"[Warning] {self.objectName}: " + text

    def formatOKString(self, text: str) -> str:
        return f"[OK] {self.objectName}: " + text

    @staticmethod
    def reportSubObjects(
        reporters: Sequence[BaseCompatibilityReporter],
        showOK: bool = True,
        showWarnings: bool = True,
    ) -> list[str]:
        report = []
        for reporter in reporters:
            if showOK or reporter.fatal or (showWarnings and reporter.warning):
                report.append(repr(reporter))
        return report

    @staticmethod
    def reportCountDifference(
        subObjectName: str,
        object1Name: str,
        object1Count: int,
        object2Name: str,
        object2Count: int,
    ) -> str:
        text = f"{object1Name} contains {object1Count} {subObjectName} | {object2Name} contains {object2Count} {subObjectName}"
        return text

    @staticmethod
    def reportOrderDifference(
        subObjectName: str,
        object1Name: str,
        object1Order: list[str],
        object2Name: str,
        object2Order: list[str],
    ) -> str:
        text = f"{object1Name} has {subObjectName} ordered {object1Order} | {object2Name} has {object2Order}"
        return text

    @staticmethod
    def reportDifferences(
        object1Name: str, subObjectName: str, subObjectID: str, object2Name: str
    ) -> str:
        text = (
            f"{object1Name} contains {subObjectName} {subObjectID} not in {object2Name}"
        )
        return text


# ----
# Font
# ----


class FontCompatibilityReporter(BaseCompatibilityReporter):
    objectName = "Font"

    def __init__(self, font1: BaseFont, font2: BaseFont) -> None:
        super().__init__(font1, font2)
        self.guidelineCountDifference = False
        self.layerCountDifference = False
        self.guidelinesMissingFromFont2: list[str] = []
        self.guidelinesMissingInFont1: list[str] = []
        self.layersMissingFromFont2: list[str] = []
        self.layersMissingInFont1: list[str] = []
        self.layers: list[LayerCompatibilityReporter] = []

    font1 = dynamicProperty("object1")
    font1Name = dynamicProperty("object1Name")
    font2 = dynamicProperty("object2")
    font2Name = dynamicProperty("object2Name")

    def report(self, showOK: bool = True, showWarnings: bool = True) -> str:
        font1 = self.font1
        font2 = self.font2
        report = []
        if self.guidelineCountDifference:
            text = self.reportCountDifference(
                subObjectName="guidelines",
                object1Name=self.font1Name,
                object1Count=len(font1.guidelines),
                object2Name=self.font2Name,
                object2Count=len(font2.guidelines),
            )
            report.append(self.formatWarningString(text))
        for name in self.guidelinesMissingFromFont2:
            text = self.reportDifferences(
                object1Name=self.font1Name,
                subObjectName="guideline",
                subObjectID=name,
                object2Name=self.font2Name,
            )
            report.append(self.formatWarningString(text))
        for name in self.guidelinesMissingInFont1:
            text = self.reportDifferences(
                object1Name=self.font2Name,
                subObjectName="guideline",
                subObjectID=name,
                object2Name=self.font1Name,
            )
            report.append(self.formatWarningString(text))
        if self.layerCountDifference:
            text = self.reportCountDifference(
                subObjectName="layers",
                object1Name=self.font1Name,
                object1Count=len(font1.layerOrder),
                object2Name=self.font2Name,
                object2Count=len(font2.layerOrder),
            )
            report.append(self.formatWarningString(text))
        for name in self.layersMissingFromFont2:
            text = self.reportDifferences(
                object1Name=self.font1Name,
                subObjectName="layer",
                subObjectID=name,
                object2Name=self.font2Name,
            )
            report.append(self.formatWarningString(text))
        for name in self.layersMissingInFont1:
            text = self.reportDifferences(
                object1Name=self.font2Name,
                subObjectName="layer",
                subObjectID=name,
                object2Name=self.font1Name,
            )
            report.append(self.formatWarningString(text))
        report += self.reportSubObjects(
            self.layers, showOK=showOK, showWarnings=showWarnings
        )

        if report or showOK:
            report.insert(0, self.title)
        return "\n".join(report)


# -----
# Layer
# -----


class LayerCompatibilityReporter(BaseCompatibilityReporter):
    objectName = "Layer"

    def __init__(self, layer1: BaseLayer, layer2: BaseLayer) -> None:
        super().__init__(layer1, layer2)
        self.glyphCountDifference = False
        self.glyphsMissingFromLayer2: list[str] = []
        self.glyphsMissingInLayer1: list[str] = []
        self.glyphs: list[GlyphCompatibilityReporter] = []

    layer1 = dynamicProperty("object1")
    layer1Name = dynamicProperty("object1Name")
    layer2 = dynamicProperty("object2")
    layer2Name = dynamicProperty("object2Name")

    def report(self, showOK: bool = True, showWarnings: bool = True) -> str:
        layer1 = self.layer1
        layer2 = self.layer2
        report = []
        if self.glyphCountDifference:
            text = self.reportCountDifference(
                subObjectName="glyphs",
                object1Name=self.layer1Name,
                object1Count=len(layer1),
                object2Name=self.layer2Name,
                object2Count=len(layer2),
            )
            report.append(self.formatWarningString(text))
        for name in self.glyphsMissingFromLayer2:
            text = self.reportDifferences(
                object1Name=self.layer1Name,
                subObjectName="glyph",
                subObjectID=name,
                object2Name=self.layer2Name,
            )
            report.append(self.formatWarningString(text))
        for name in self.glyphsMissingInLayer1:
            text = self.reportDifferences(
                object1Name=self.layer2Name,
                subObjectName="glyph",
                subObjectID=name,
                object2Name=self.layer1Name,
            )
            report.append(self.formatWarningString(text))
        report += self.reportSubObjects(
            self.glyphs, showOK=showOK, showWarnings=showWarnings
        )

        if report or showOK:
            report.insert(0, self.title)
        return "\n".join(report)


# -----
# Glyph
# -----


class GlyphCompatibilityReporter(BaseCompatibilityReporter):
    objectName = "Glyph"

    def __init__(self, glyph1: BaseGlyph, glyph2: BaseGlyph) -> None:
        super().__init__(glyph1, glyph2)
        self.contourCountDifference = False
        self.componentCountDifference = False
        self.guidelineCountDifference = False
        self.anchorDifferences: DiffType = []
        self.anchorCountDifference = False
        self.anchorOrderDifference = False
        self.anchorsMissingFromGlyph1: list[str] = []
        self.anchorsMissingFromGlyph2: list[str] = []
        self.componentDifferences: DiffType = []
        self.componentOrderDifference = False
        self.componentsMissingFromGlyph1: list[str] = []
        self.componentsMissingFromGlyph2: list[str] = []
        self.guidelinesMissingFromGlyph1: list[str] = []
        self.guidelinesMissingFromGlyph2: list[str] = []
        self.contours: list[ContourCompatibilityReporter] = []

    glyph1 = dynamicProperty("object1")
    glyph1Name = dynamicProperty("object1Name")
    glyph2 = dynamicProperty("object2")
    glyph2Name = dynamicProperty("object2Name")

    def report(self, showOK: bool = True, showWarnings: bool = True) -> str:
        glyph1 = self.glyph1
        glyph2 = self.glyph2
        report = []

        # Contour test
        if self.contourCountDifference:
            text = self.reportCountDifference(
                subObjectName="contours",
                object1Name=self.glyph1Name,
                object1Count=len(glyph1),
                object2Name=self.glyph2Name,
                object2Count=len(glyph2),
            )
            report.append(self.formatFatalString(text))
        report += self.reportSubObjects(
            self.contours, showOK=showOK, showWarnings=showWarnings
        )

        # Component test
        if self.componentCountDifference:
            text = self.reportCountDifference(
                subObjectName="components",
                object1Name=self.glyph1Name,
                object1Count=len(glyph1.components),
                object2Name=self.glyph2Name,
                object2Count=len(glyph2.components),
            )
            report.append(self.formatFatalString(text))
        elif self.componentOrderDifference:
            text = self.reportOrderDifference(
                subObjectName="components",
                object1Name=self.glyph1Name,
                object1Order=[c.baseGlyph for c in glyph1.components],
                object2Name=self.glyph2Name,
                object2Order=[c.baseGlyph for c in glyph2.components],
            )
            report.append(self.formatWarningString(text))
        for name in self.componentsMissingFromGlyph2:
            text = self.reportDifferences(
                object1Name=self.glyph1Name,
                subObjectName="component",
                subObjectID=name,
                object2Name=self.glyph2Name,
            )
            report.append(self.formatWarningString(text))
        for name in self.componentsMissingFromGlyph1:
            text = self.reportDifferences(
                object1Name=self.glyph2Name,
                subObjectName="component",
                subObjectID=name,
                object2Name=self.glyph1Name,
            )
            report.append(self.formatWarningString(text))

        # Anchor test
        if self.anchorCountDifference:
            text = self.reportCountDifference(
                subObjectName="anchors",
                object1Name=self.glyph1Name,
                object1Count=len(glyph1.anchors),
                object2Name=self.glyph2Name,
                object2Count=len(glyph2.anchors),
            )
            report.append(self.formatWarningString(text))
        elif self.anchorOrderDifference:
            text = self.reportOrderDifference(
                subObjectName="anchors",
                object1Name=self.glyph1Name,
                object1Order=[a.name for a in glyph1.anchors],
                object2Name=self.glyph2Name,
                object2Order=[a.name for a in glyph2.anchors],
            )
            report.append(self.formatWarningString(text))
        for name in self.anchorsMissingFromGlyph2:
            text = self.reportDifferences(
                object1Name=self.glyph1Name,
                subObjectName="anchor",
                subObjectID=name,
                object2Name=self.glyph2Name,
            )
            report.append(self.formatWarningString(text))
        for name in self.anchorsMissingFromGlyph1:
            text = self.reportDifferences(
                object1Name=self.glyph2Name,
                subObjectName="anchor",
                subObjectID=name,
                object2Name=self.glyph1Name,
            )
            report.append(self.formatWarningString(text))

        # Guideline test
        if self.guidelineCountDifference:
            text = self.reportCountDifference(
                subObjectName="guidelines",
                object1Name=self.glyph1Name,
                object1Count=len(glyph1.guidelines),
                object2Name=self.glyph2Name,
                object2Count=len(glyph2.guidelines),
            )
            report.append(self.formatWarningString(text))
        for name in self.guidelinesMissingFromGlyph2:
            text = self.reportDifferences(
                object1Name=self.glyph1Name,
                subObjectName="guideline",
                subObjectID=name,
                object2Name=self.glyph2Name,
            )
            report.append(self.formatWarningString(text))
        for name in self.guidelinesMissingFromGlyph1:
            text = self.reportDifferences(
                object1Name=self.glyph2Name,
                subObjectName="guideline",
                subObjectID=name,
                object2Name=self.glyph1Name,
            )
            report.append(self.formatWarningString(text))

        if report or showOK:
            report.insert(0, self.title)
        return "\n".join(report)


# -------
# Contour
# -------


class ContourCompatibilityReporter(BaseCompatibilityReporter):
    objectName = "Contour"

    def __init__(self, contour1: BaseContour, contour2: BaseContour) -> None:
        super().__init__(contour1, contour2)
        self.openDifference = False
        self.directionDifference = False
        self.segmentCountDifference = False
        self.segments: list[SegmentCompatibilityReporter] = []

    contour1 = dynamicProperty("object1")
    contour1Name = dynamicProperty("object1Name")
    contour2 = dynamicProperty("object2")
    contour2Name = dynamicProperty("object2Name")

    def report(self, showOK: bool = True, showWarnings: bool = True) -> str:
        contour1 = self.contour1
        contour2 = self.contour2
        report = []
        if self.segmentCountDifference:
            text = self.reportCountDifference(
                subObjectName="segments",
                object1Name=self.contour1Name,
                object1Count=len(contour1),
                object2Name=self.contour2Name,
                object2Count=len(contour2),
            )
            report.append(self.formatFatalString(text))
        if self.openDifference:
            state1 = state2 = "closed"
            if contour1.open:
                state1 = "open"
            if contour2.open:
                state2 = "open"
            text = f"{self.contour1Name} is {state1} | {self.contour2Name} is {state2}"
            report.append(self.formatFatalString(text))
        if self.directionDifference:
            state1 = state2 = "counter-clockwise"
            if contour1.clockwise:
                state1 = "clockwise"
            if contour2.clockwise:
                state2 = "clockwise"
            text = f"{self.contour1Name} is {state1} | {self.contour2Name} is {state2}"
            report.append(self.formatFatalString(text))
        report += self.reportSubObjects(
            self.segments, showOK=showOK, showWarnings=showWarnings
        )
        if report or showOK:
            report.insert(0, self.title)
        return "\n".join(report)


# -------
# Segment
# -------


class SegmentCompatibilityReporter(BaseCompatibilityReporter):
    objectName = "Segment"

    def __init__(self, contour1: BaseContour, contour2: BaseContour) -> None:
        super().__init__(contour1, contour2)
        self.typeDifference = False

    segment1 = dynamicProperty("object1")
    segment1Name = dynamicProperty("object1Name")
    segment2 = dynamicProperty("object2")
    segment2Name = dynamicProperty("object2Name")

    def report(self, showOK: bool = True, showWarnings: bool = True) -> str:
        segment1 = self.segment1
        segment2 = self.segment2
        report = []
        if self.typeDifference:
            type1 = segment1.type
            type2 = segment2.type
            text = f"{self.segment1Name} is {type1} | {self.segment2Name} is {type2}"
            report.append(self.formatFatalString(text))
        if report or showOK:
            report.insert(0, self.title)
        return "\n".join(report)


# ---------
# Component
# ---------


class ComponentCompatibilityReporter(BaseCompatibilityReporter):
    objectName = "Component"

    def __init__(self, component1: BaseComponent, component2: BaseComponent) -> None:
        super().__init__(component1, component2)
        self.baseDifference = False

    component1 = dynamicProperty("object1")
    component1Name = dynamicProperty("object1Name")
    component2 = dynamicProperty("object2")
    component2Name = dynamicProperty("object2Name")

    def report(self, showOK: bool = True, showWarnings: bool = True) -> str:
        component1 = self.component1
        component2 = self.component2
        report = []
        if self.baseDifference:
            name1 = component1.baseName
            name2 = component2.baseName
            text = f"{self.component1Name} has base glyph {name1} | {self.component2Name} has base glyph {name2}"
            report.append(self.formatWarningString(text))
        if report or showOK:
            report.insert(0, self.title)
        return "\n".join(report)


# ------
# Anchor
# ------


class AnchorCompatibilityReporter(BaseCompatibilityReporter):
    objectName = "Anchor"

    def __init__(self, anchor1: BaseAnchor, anchor2: BaseAnchor) -> None:
        super().__init__(anchor1, anchor2)
        self.nameDifference = False

    anchor1 = dynamicProperty("object1")
    anchor1Name = dynamicProperty("object1Name")
    anchor2 = dynamicProperty("object2")
    anchor2Name = dynamicProperty("object2Name")

    def report(self, showOK: bool = True, showWarnings: bool = True) -> str:
        anchor1 = self.anchor1
        anchor2 = self.anchor2
        report = []
        if self.nameDifference:
            name1 = anchor1.name
            name2 = anchor2.name
            text = f"{self.anchor1Name} has name {name1} | {self.anchor2Name} has name {name2}"
            report.append(self.formatWarningString(text))
        if report or showOK:
            report.insert(0, self.title)
        return "\n".join(report)


# ---------
# Guideline
# ---------


class GuidelineCompatibilityReporter(BaseCompatibilityReporter):
    objectName = "Guideline"

    def __init__(self, guideline1: BaseGuideline, guideline2: BaseGuideline) -> None:
        super().__init__(guideline1, guideline2)
        self.nameDifference = False

    guideline1 = dynamicProperty("object1")
    guideline1Name = dynamicProperty("object1Name")
    guideline2 = dynamicProperty("object2")
    guideline2Name = dynamicProperty("object2Name")

    def report(self, showOK: bool = True, showWarnings: bool = True) -> str:
        guideline1 = self.guideline1
        guideline2 = self.guideline2
        report = []
        if self.nameDifference:
            name1 = guideline1.name
            name2 = guideline2.name
            text = f"{self.guideline1Name} has name {name1} | {self.guideline2Name} has name {name2}"
            report.append(self.formatWarningString(text))
        if report or showOK:
            report.insert(0, self.title)
        return "\n".join(report)



from __future__ import annotations
from typing import TYPE_CHECKING, Any

import defcon
from fontParts.base import BaseLayer
from fontParts.base.annotations import RGBALike, RGBA
from fontParts.fontshell.base import RBaseObject
from fontParts.fontshell.lib import RLib
from fontParts.fontshell.glyph import RGlyph

if TYPE_CHECKING:
    from fontParts.base.glyph import BaseGlyph


class RLayer(RBaseObject, BaseLayer):
    wrapClass = defcon.Layer
    libClass = RLib
    glyphClass = RGlyph

    # -----------
    # Sub-Objects
    # -----------

    # lib

    def _get_lib(self) -> RLib:
        return self.libClass(pathOrObject=self.naked().lib)

    # tempLib

    def _get_tempLib(self) -> RLib:
        return self.libClass(pathOrObject=self.naked().tempLib)

    # --------------
    # Identification
    # --------------

    # name

    def _get_name(self) -> str:
        return self.naked().name

    def _set_name(self, value: str, **kwargs: Any) -> None:
        self.naked().name = value

    # color

    def _get_color(self) -> RGBA | None:
        value = self.naked().color
        if value is not None:
            value = tuple(value)
        return value

    def _set_color(self, value: RGBALike | None, **kwargs: Any) -> None:
        self.naked().color = value

    # -----------------
    # Glyph Interaction
    # -----------------

    def _getItem(self, name: str, **kwargs: Any) -> RGlyph:
        layer = self.naked()
        glyph = layer[name]
        return self.glyphClass(glyph)

    def _keys(self, **kwargs: Any) -> tuple[str, ...]:
        return tuple(self.naked().keys())

    def _newGlyph(self, name: str, **kwargs: Any) -> BaseGlyph:
        layer = self.naked()
        layer.newGlyph(name)
        return self[name]

    def _removeGlyph(self, name: str, **kwargs: Any) -> None:
        layer = self.naked()
        del layer[name]

    # -------
    # mapping
    # -------

    def _getReverseComponentMapping(self) -> dict[str, tuple[str, ...]]:
        mapping = self.naked().componentReferences
        return {k: tuple(v) for k, v in mapping.items()}

    def _getCharacterMapping(self) -> dict[int, tuple[str, ...]]:
        mapping = self.naked().unicodeData
        return {k: tuple(v) for k, v in mapping.items()}



import unittest
import collections


class TestInfo(unittest.TestCase):
    def getInfo_generic(self):
        info, _ = self.objectGenerator("info")
        info.unitsPerEm = 1000
        return info

    # ----------
    # Dimensions
    # ----------

    def test_get_unitsPerEm(self):
        info = self.getInfo_generic()
        self.assertEqual(info.unitsPerEm, 1000)

    def test_set_valid_unitsPerEm_int(self):
        info = self.getInfo_generic()
        info.unitsPerEm = 2000
        self.assertEqual(info.unitsPerEm, 2000)

    def test_set_valid_unitsPerEm_float(self):
        info = self.getInfo_generic()
        info.unitsPerEm = 2000.1
        self.assertEqual(info.unitsPerEm, 2000.1)

    def test_set_invalid_unitsPerEm_negative(self):
        info = self.getInfo_generic()
        with self.assertRaises(ValueError):
            info.unitsPerEm = -1000

    def test_set_invalid_unitsPerEm_string(self):
        info = self.getInfo_generic()
        with self.assertRaises(ValueError):
            info.unitsPerEm = "abc"

    # ----
    # Hash
    # ----

    def test_hash(self):
        info = self.getInfo_generic()
        self.assertEqual(isinstance(info, collections.abc.Hashable), True)

    # --------
    # Equality
    # --------

    def test_object_equal_self(self):
        info_one = self.getInfo_generic()
        self.assertEqual(info_one, info_one)

    def test_object_not_equal_other(self):
        info_one = self.getInfo_generic()
        info_two = self.getInfo_generic()
        self.assertNotEqual(info_one, info_two)

    def test_object_equal_self_variable_assignment(self):
        info_one = self.getInfo_generic()
        a = info_one
        self.assertEqual(info_one, a)

    def test_object_not_equal_other_variable_assignment(self):
        info_one = self.getInfo_generic()
        info_two = self.getInfo_generic()
        a = info_one
        self.assertNotEqual(info_two, a)

    # -----
    # Round
    # -----

    def test_round_unitsPerEm(self):
        info = self.getInfo_generic()
        info.unitsPerEm = 2000.125
        info.round()
        self.assertEqual(info.unitsPerEm, 2000)

    # ------
    # Update
    # ------

    def test_update(self):

        info1 = self.getInfo_generic()
        info1.familyName = "test1"
        info1.unitsPerEm = 1000
        info2 = self.getInfo_generic()
        info2.familyName = "test2"
        info2.unitsPerEm = 2000
        info1.update(info2)
        self.assertEqual(info1.familyName, "test2")
        self.assertEqual(info1.unitsPerEm, 2000)

    # ----
    # Copy
    # ----
    def test_copy(self):
        info1 = self.getInfo_generic()
        info1.postscriptBlueValues = [-10, 0, 50, 60]

        info2 = info1.copy()
        info2.postscriptBlueValues[0] = -2

        self.assertNotEqual(info1.postscriptBlueValues, info2.postscriptBlueValues)

    # -------------
    # Interpolation
    # -------------

    def test_interpolate_unitsPerEm_without_rounding(self):
        interpolated_font, _ = self.objectGenerator("font")
        font_min, _ = self.objectGenerator("font")
        font_max, _ = self.objectGenerator("font")
        font_min.info.unitsPerEm = 1000
        font_max.info.unitsPerEm = 2000
        interpolated_font.info.interpolate(
            0.5154, font_min.info, font_max.info, round=False
        )
        self.assertEqual(interpolated_font.info.unitsPerEm, 1515.4)

    def test_interpolate_unitsPerEm_with_rounding(self):
        interpolated_font, _ = self.objectGenerator("font")
        font_min, _ = self.objectGenerator("font")
        font_max, _ = self.objectGenerator("font")
        font_min.info.unitsPerEm = 1000
        font_max.info.unitsPerEm = 2000
        interpolated_font.info.interpolate(
            0.5154, font_min.info, font_max.info, round=True
        )
        self.assertEqual(interpolated_font.info.unitsPerEm, 1515)



from __future__ import annotations

import defcon
from fontTools.ufoLib.validators import pngValidator
from fontParts.base import BaseImage, FontPartsError
from fontParts.base.annotations import (
    AffineTransformationLike,
    AffineTransformation,
    RGBALike,
    RGBA,
)
from fontParts.fontshell.base import RBaseObject


class RImage(RBaseObject, BaseImage):
    wrapClass = defcon.Image
    _orphanData: bytes | None = None
    _orphanColor: RGBALike | None = None

    # ----------
    # Attributes
    # ----------

    # Transformation

    def _get_transformation(self) -> AffineTransformation:
        return self.naked().transformation

    def _set_transformation(self, value: AffineTransformationLike) -> None:
        self.naked().transformation = value

    # Color

    def _get_color(self) -> RGBA | None:
        if self.font is None and self._orphanColor is not None:
            r, g, b, a = self._orphanColor
            return (r, g, b, a)
        value = self.naked().color
        if value is not None:
            value = tuple(value)
        return value

    def _set_color(self, value: RGBALike | None) -> None:
        if self.font is None:
            self._orphanColor = value
        else:
            self.naked().color = value

    # Data

    def _get_data(self) -> bytes | None:
        if self.font is None:
            return self._orphanData
        image = self.naked()
        images = self.font.naked().images
        fileName = image.fileName
        if fileName is None:
            return None
        if fileName not in images:
            return None
        return images[fileName]

    def _set_data(self, value: bytes) -> None:
        if not isinstance(value, bytes):
            raise FontPartsError("The image data provided is not valid.")
        if not pngValidator(data=value)[0]:
            raise FontPartsError("The image must be in PNG format.")
        if self.font is None:
            self._orphanData = value
        else:
            image = self.naked()
            images = image.font.images
            fileName = images.findDuplicateImage(value)
            if fileName is None:
                fileName = images.makeFileName("image.png")
                images[fileName] = value
            image.fileName = fileName



from __future__ import annotations
from typing import TYPE_CHECKING
from collections.abc import Callable

from fontParts.base.base import BaseObject, dynamicProperty, reference
from fontParts.base import normalizers
from fontParts.base.deprecated import DeprecatedFeatures, RemovedFeatures

if TYPE_CHECKING:
    from fontParts.base.font import BaseFont


class BaseFeatures(BaseObject, DeprecatedFeatures, RemovedFeatures):
    """Represent the basis for a features object.

    This class contains the font's `Adobe Font Development Kit for OpenType (AFDKO)
    <https://github.com/adobe-type-tools/afdko/>`_ feature definitions.

    """

    copyAttributes: tuple[str] = ("text",)

    def _reprContents(self) -> list[str]:
        contents = []
        if self.font is not None:
            contents.append("for font")
            contents += self.font._reprContents()
        return contents

    # -------
    # Parents
    # -------

    # Font

    _font: Callable[[], BaseFont] | None = None

    font: dynamicProperty = dynamicProperty(
        "font",
        """Get the feature's parent font object.

        This property is read-only.

        :return: The :class:`BaseFont` instance containing the features
            or :obj:`None`.
        :raises AssertionError:
            - If attempting to set the font when it has already been set and is
              not the same as the provided font.

        Example::

            >>> font = features.font

        """,
    )

    def _get_font(self) -> BaseFont | None:
        if self._font is None:
            return None
        return self._font()

    def _set_font(self, font: BaseFont | Callable[[], BaseFont] | None) -> None:
        if self._font is not None and self._font() != font:
            raise AssertionError(
                "font for features already set and is not same as font"
            )
        if font is not None:
            font = reference(font)
        self._font = font

    # ----
    # Text
    # ----

    text: dynamicProperty = dynamicProperty(
        "base_text",
        """Get or set the features text.
        
        The value must be a `AFDKO FEA formatted
        <https://adobe-type-tools.github.io/afdko/OpenTypeFeatureFileSpecification.html>`_
        :class:`str` or :obj:`None`.
        
        :return: A :class:`str` representing the font's features text, or :obj:`None`.

        """,
    )

    def _get_base_text(self) -> str | None:
        value = self._get_text()
        if value is not None:
            value = normalizers.normalizeFeatureText(value)
        return value

    def _set_base_text(self, value: str | None) -> None:
        if value is not None:
            value = normalizers.normalizeFeatureText(value)
        self._set_text(value)

    def _get_text(self) -> str | None:
        """Get the features text.

        This is the environment implementation of the :attr:`BaseFeatures.text` property
        getter.

        :return: :class:`_empty`. The value will be normalized with
            :func:`normalizers.normalizeFeatureText`.
        :raises NotImplementedError: If the method has not been overridden by a
            subclass.

        .. important::

            Subclasses must override this method.

        """
        self.raiseNotImplementedError()

    def _set_text(self, value: str | None) -> None:
        """Set the features text.

        Description

        This is the environment implementation of the :attr:`BaseFeatures.text` property
        setter.

        :param value: A `AFDKO FEA formatted
        <https://adobe-type-tools.github.io/afdko/OpenTypeFeatureFileSpecification.html>`_
        :class:`str` or :obj:`None`. The value will have been
            normalized with :func:`normalizers.normalizeFeatureText`.
        :raises NotImplementedError: If the method has not been overridden by a
            subclass.

        .. important::

            Subclasses must override this method.

        """
        self.raiseNotImplementedError()



from __future__ import annotations
from typing import Generic, TypeVar

RBaseObjectType = TypeVar("RBaseObjectType", bound="RBaseObject")


class RBaseObject(Generic[RBaseObjectType]):
    wrapClass: type[RBaseObjectType] | None = None
    dirty: bool

    def _init(self, pathOrObject: RBaseObjectType | None = None) -> None:
        if pathOrObject is None and self.wrapClass is not None:
            pathOrObject = self.wrapClass()  # pylint: disable=E1102
        if pathOrObject is not None:
            self._wrapped = pathOrObject

    def changed(self) -> None:
        self.naked().dirty = True

    def naked(self) -> RBaseObjectType:
        if hasattr(self, "_wrapped"):
            return self._wrapped
        return None  # type: ignore[return-value]



from __future__ import annotations
from typing import TYPE_CHECKING, Any
from collections.abc import Callable

from fontTools.misc import transform
from fontTools.pens.pointInsidePen import PointInsidePen
from fontTools.pens.boundsPen import BoundsPen
from fontParts.base import normalizers
from fontParts.base.errors import FontPartsError
from fontParts.base.base import (
    BaseObject,
    TransformationMixin,
    InterpolationMixin,
    SelectionMixin,
    IdentifierMixin,
    dynamicProperty,
    reference,
)
from fontParts.base.bounds import Bounds
from fontParts.base.compatibility import ComponentCompatibilityReporter
from fontParts.base.deprecated import DeprecatedComponent, RemovedComponent
from fontParts.base.annotations import (
    ScaleFactorPair,
    ScaleFactor,
    AffineTransformationLike,
    AffineTransformation,
    BoundingBox,
    Coordinate,
    CoordinateLike,
    PenType,
    PointPenType,
)

if TYPE_CHECKING:
    from fontParts.base.glyph import BaseGlyph
    from fontParts.base.layer import BaseLayer
    from fontParts.base.font import BaseFont


class BaseComponent(
    BaseObject,
    TransformationMixin,
    InterpolationMixin,
    SelectionMixin,
    IdentifierMixin,
    DeprecatedComponent,
    RemovedComponent,
):
    """Represent the basis for a component object.

    This object provides a reference to another glyph, allowing it to be
    inserted as part of an outline.

    """

    copyAttributes: tuple[str, str] = ("baseGlyph", "transformation")

    def _reprContents(self) -> list[str]:
        contents = [
            f"baseGlyph='{self.baseGlyph}'",
            f"offset='({self.offset[0]}, {self.offset[1]})'",
        ]
        if self.glyph is not None:
            contents.append("in glyph")
            contents += self.glyph._reprContents()
        return contents

    # -------
    # Parents
    # -------

    # Glyph

    _glyph: Callable[[], BaseGlyph] | None = None

    glyph: dynamicProperty = dynamicProperty(
        "glyph",
        """Get or set the component's parent glyph object.

        The value must be a :class:`BaseGlyph` instance or :obj:`None`.

        :return: The :class:`BaseGlyph` instance containing the component
            or :obj:`None`.
        :raises AssertionError: If attempting to set the glyph when it
            has already been set.

        Example::

            >>> glyph = component.glyph

        """,
    )

    def _get_glyph(self) -> BaseGlyph | None:
        if self._glyph is None:
            return None
        return self._glyph()

    def _set_glyph(self, glyph: BaseGlyph | Callable[[], BaseGlyph] | None) -> None:
        if self._glyph is not None:
            raise AssertionError("glyph for component already set")
        if glyph is not None:
            glyph = reference(glyph)
        self._glyph = glyph

    # Layer

    layer: dynamicProperty = dynamicProperty(
        "layer",
        """Get the component's parent layer object.

        This property is read-only.

        :return: The :class:`BaseLayer` instance containing the component
            or :obj:`None`.

        Example::

            >>> layer = component.layer

        """,
    )

    def _get_layer(self) -> BaseLayer | None:
        if self._glyph is None:
            return None
        return self.glyph.layer

    # Font

    font: dynamicProperty = dynamicProperty(
        "font",
        """Get the component's parent font object.

        This property is read-only.

        :return: The :class:`BaseFont` instance containing the component
            or :obj:`None`.

        Example::

            >>> font = component.font

        """,
    )

    def _get_font(self) -> BaseFont | None:
        if self._glyph is None:
            return None
        return self.glyph.font

    # ----------
    # Attributes
    # ----------

    # baseGlyph

    baseGlyph: dynamicProperty = dynamicProperty(
        "base_baseGlyph",
        """Get or set the name of the glyph referenced by the component.

        The value must be a :class:`str`.

        :return: A :class:`str` representing the name of the base glyph,
            or :obj:`None` if the component does not belong to a layer.
        :raise ValueError: If value is None when the component is part of a layer.

        """,
    )

    def _get_base_baseGlyph(self) -> str | None:
        value = self._get_baseGlyph()
        # if the component does not belong to a layer,
        # it is allowed to have None as its baseGlyph
        if value is None and self.layer is None:
            return value
        if value is None:
            raise ValueError(f"Value cannot be None when layer is '{self.layer}'.")
        return normalizers.normalizeGlyphName(value)

    def _set_base_baseGlyph(self, value: str) -> None:
        value = normalizers.normalizeGlyphName(value)
        self._set_baseGlyph(value)

    def _get_baseGlyph(self) -> str | None:
        """Get the name of the glyph referenced by the native component.

        This is the environment implementation of the :attr:`BaseComponent.baseGlyph`
        property getter.

        :return: A :class:`str` representing the name of the base glyph,
            or :obj:`None` if the component does not belong to a layer. The value
            will be normalized with :func:`normalizers.normalizeGlyphName`.
        :raises NotImplementedError: If the method has not been overridden by a
            subclass.

        .. important::

            Subclasses must override this method.

        """
        self.raiseNotImplementedError()

    def _set_baseGlyph(self, value: str) -> None:
        """Set the name of the glyph referenced by the native component.

        This is the environment implementation of the :attr:`BaseComponent.baseGlyph`
        property setter.

        :param value: The name of the glyph to set as a :class:`str`. The value
            will have been normalized
         with :func:`normalizers.normalizeGlyphName`.
        :raises NotImplementedError: If the method has not been overridden by a
            subclass.

        .. important::

            Subclasses must override this method.

        """
        self.raiseNotImplementedError()

    # transformation

    transformation: dynamicProperty = dynamicProperty(
        "base_transformation",
        """Get or set the component's transformation matrix.

        The value must be a :ref:`type-transformation`.

        :return: A :ref:`type-transformation` value representing the
            transformation matrix of the component.

        """,
    )

    def _get_base_transformation(self) -> AffineTransformation:
        value = self._get_transformation()
        value = normalizers.normalizeTransformationMatrix(value)
        return value

    def _set_base_transformation(self, value: AffineTransformationLike) -> None:
        value = normalizers.normalizeTransformationMatrix(value)
        self._set_transformation(value)

    def _get_transformation(self) -> AffineTransformation:
        """Get the native component's transformation matrix.

        This is the environment implementation of the
        :attr:`BaseComponent.transformation` property getter.

        :return: A :ref:`type-transformation` value representing the
            transformation matrix of the component. The value will be
            normalized with :func:`normalizers.normalizeTransformationMatrix`.
        :raises NotImplementedError: If the method has not been overridden by a
            subclass.

        .. important::

            Subclasses must override this method.

        """
        self.raiseNotImplementedError()

    def _set_transformation(self, value: AffineTransformationLike) -> None:
        """Set the native component's transformation matrix.

        This is the environment implementation of the
        :attr:`BaseComponent.transformation` property setter.

        :param value: The :ref:`type-transformation` to set. The value will have
            been normalized with :func:`normalizers.normalizeTransformationMatrix`.
        :raises NotImplementedError: If the method has not been overridden by a
            subclass.

        .. important::

            Subclasses must override this method.

        """
        self.raiseNotImplementedError()

    # offset

    offset: dynamicProperty = dynamicProperty(
        "base_offset",
        """Get or set the component's offset.

        The value must be a :ref:`type-coordinate`.

        :return: A :ref:`type-coordinate` representing the offset of the component.

        """,
    )

    def _get_base_offset(self) -> Coordinate:
        value = self._get_offset()
        value = normalizers.normalizeTransformationOffset(value)
        return value

    def _set_base_offset(self, value: CoordinateLike) -> None:
        value = normalizers.normalizeTransformationOffset(value)
        self._set_offset(value)

    def _get_offset(self) -> Coordinate:
        """Get the native component's offset.

        This is the environment implementation of the :attr:`BaseComponent.offset`
        property getter.

        :return: A :ref:`type-coordinate` representing the offset of the component.
            The value will be normalized
            with :func:`normalizers.normalizeTransformationOffset`.

        .. note::

            Subclasses may override this method.

        """
        sx, sxy, syx, sy, ox, oy = self.transformation
        return ox, oy

    def _set_offset(self, value: CoordinateLike) -> None:
        """Set the native component's offset.

        This is the environment implementation of the :attr:`BaseComponent.offset`
        property setter.

        :param value: The offset to set as a :ref:`type-coordinate`. The value will
            have been normalized with :func:`normalizers.normalizeTransformationOffset`.

        .. note::

            Subclasses may override this method.

        """
        sx, sxy, syx, sy, ox, oy = self.transformation
        ox, oy = value
        self.transformation = sx, sxy, syx, sy, ox, oy

    # scale

    scale: dynamicProperty = dynamicProperty(
        "base_scale",
        """Get or set the component's scale.

        The value must be a :class:`list` or :class:`tuple` of two :class:`int`
        or :class:`float` items representing the ``(x, y)`` scale of the component.

        :return: A :class:`tuple` of two :class:`float` items representing the
            ``(x, y)`` scale of the component.

        """,
    )

    def _get_base_scale(self) -> ScaleFactor:
        value = self._get_scale()
        value = normalizers.normalizeComponentScale(value)
        return value

    def _set_base_scale(self, value: ScaleFactorPair) -> None:
        value = normalizers.normalizeComponentScale(value)
        self._set_scale(value)

    def _get_scale(self) -> ScaleFactor:
        """Get the native component's scale.

        This is the environment implementation of the :attr:`BaseComponent.scale`
        property getter.

        :return: A :class:`tuple` of two :class:`float` items representing the
            ``(x, y)`` scale of the component. The value will have been normalized
            with :func:`normalizers.normalizeComponentScale`.

        .. note::

            Subclasses may override this method.

        """
        sx, sxy, syx, sy, ox, oy = self.transformation
        return sx, sy

    def _set_scale(self, value: ScaleFactorPair) -> None:
        """Set the native component's scale.

        This is the environment implementation of the :attr:`BaseComponent.scale`
        property setter.

        :param value: The scale to set as a :class:`list` or :class:`tuple`
            of :class:`int` or :class:`float` items representing the ``(x, y)``
            scale of the component. The value will have been normalized
            with :func:`normalizers.normalizeComponentScale`.

        .. note::

            Subclasses may override this method.

        """
        sx, sxy, syx, sy, ox, oy = self.transformation
        sx, sy = value
        self.transformation = sx, sxy, syx, sy, ox, oy

    # --------------
    # Identification
    # --------------

    # index

    index: dynamicProperty = dynamicProperty(
        "base_index",
        """Get or set the index of the contour.

        The value must be an :class:`int`

        :return: An :class:`int` representing the index of the component within
            the ordered list of parent glyph's components, or :obj:`None` if the
            component does not belong to a glyph.
        :raise FontPartsError: If attempting to set the index while the
            component does not belong to a glyph.

        """,
    )

    def _get_base_index(self) -> int | None:
        glyph = self.glyph
        if glyph is None:
            return None
        value = self._get_index()
        value = normalizers.normalizeIndex(value)
        return value

    def _set_base_index(self, value: int) -> None:
        glyph = self.glyph
        if glyph is None:
            raise FontPartsError("The component does not belong to a glyph.")
        index = normalizers.normalizeIndex(value)
        if index is None:
            return
        value = index
        componentCount = len(glyph.components)
        if value < 0:
            value = -(value % componentCount)
        if value >= componentCount:
            value = componentCount
        self._set_index(value)

    def _get_index(self) -> int | None:
        """Get the index of the native contour.

        This is the environment implementation of the :attr:`BaseComponent.index`
        property getter.

        :return: An :class:`int` representing the index of the component within
            the ordered list of parent glyph's components, or :obj:`None` if the
            component does not belong to a glyph. The value will be normalized
            with :func:`normalizers.normalizeIndex`.

        .. note::

            Subclasses may override this method.

        """
        glyph = self.glyph
        return glyph.components.index(self)

    def _set_index(self, value: int) -> None:
        """Set the index of the native contour.

        This is the environment implementation of the :attr:`BaseComponent.index`
        property setter.

        :param value: The index to set as an :class:`int`. The value will have been
            normalized with :func:`normalizers.normalizeIndex`.
        :raises NotImplementedError: If the method has not been overridden by a
            subclass.

        .. important::

            Subclasses must override this method.

        """
        self.raiseNotImplementedError()

    # ----
    # Pens
    # ----

    def draw(self, pen: PenType) -> None:
        """Draw the component with the given pen.

        :param pen: The :class:`fontTools.pens.basePen.AbstractPen` with which
            to draw the componnet.

        """
        self._draw(pen)

    def _draw(self, pen: PenType, **kwargs: Any) -> None:
        r"""Draw the native component with the given pen.

        This is the environment implementation of :meth:`BaseComponent.draw`.

        :param pen: The :class:`fontTools.pens.basePen.AbstractPen` with which
            to draw the componnet.
        :param \**kwargs: Additional keyword arguments.

        .. note::

            Subclasses may override this method.

        """
        from fontTools.ufoLib.pointPen import PointToSegmentPen

        adapter = PointToSegmentPen(pen)
        self.drawPoints(adapter)

    def drawPoints(self, pen: PointPenType) -> None:
        """Draw the component with the given point pen.

        :param pen: The :class:`fontTools.pens.pointPen.AbstractPointPen` with
            which to draw the componnet.

        """
        self._drawPoints(pen)

    def _drawPoints(self, pen: PointPenType, **kwargs: Any) -> None:
        r"""Draw the native component with the given point pen.

        This is the environment implementation of :meth:`BaseComponent.draw`.

        :param pen: The :class:`fontTools.pens.pointPen.AbstractPointPen` with
            which to draw the componnet.
        :param \**kwargs: Additional keyword arguments.

        .. note::

            Subclasses may override this method.

        """
        # The try: ... except TypeError: ...
        # handles backwards compatibility with
        # point pens that have not been upgraded
        # to point pen protocol 2.
        try:
            pen.addComponent(
                self.baseGlyph,
                self.transformation,
                identifier=self.identifier,
                **kwargs,
            )
        except TypeError:
            pen.addComponent(self.baseGlyph, self.transformation, **kwargs)

    # --------------
    # Transformation
    # --------------

    def _transformBy(self, matrix: AffineTransformationLike, **kwargs: Any) -> None:
        r"""Transform the component according to the given matrix.

        This is the environment implementation of :meth:`BaseComponent.transformBy`.

        :param matrix: The :ref:`type-transformation` to apply. The value will
             be normalized with :func:`normalizers.normalizeTransformationMatrix`.
        :param \**kwargs: Additional keyword arguments.

        .. note::

            Subclasses may override this method.

        """
        t = transform.Transform(*matrix)
        transformation = t.transform(self.transformation)
        self.transformation = tuple(transformation)

    # -------------
    # Normalization
    # -------------

    def round(self) -> None:
        """Round the compnent's offset coordinates.

        This applies to :attr:`offset`

        """
        self._round()

    def _round(self) -> None:
        """Round the native compnent's offset coordinates.

        This is the environment implementation of :meth:`BaseComponent.round`.

        .. note::

            Subclasses may override this method.

        """
        x, y = self.offset
        x = normalizers.normalizeVisualRounding(x)
        y = normalizers.normalizeVisualRounding(y)
        self.offset = (x, y)

    def decompose(self) -> None:
        """Decompose the component."""
        glyph = self.glyph
        if glyph is None:
            raise FontPartsError("The component does not belong to a glyph.")
        self._decompose()

    def _decompose(self) -> None:
        """Decompose the native component.

        This is the environment implementation of :meth:`BaseComponent.decompose`.

        :raises NotImplementedError: If the method has not been overridden by a
            subclass.

        .. important::

            Subclasses must override this method.

        """
        self.raiseNotImplementedError()

    # -------------
    # Interpolation
    # -------------

    compatibilityReporterClass = ComponentCompatibilityReporter

    def isCompatible(
        self, other: BaseComponent, cls=None
    ) -> tuple[bool, ComponentCompatibilityReporter]:
        """Evaluate interpolation compatibility with another component.

        :param other: The other :class:`BaseComponent` instance to check
            compatibility with.
        :return: A :class:`tuple` where the first element is a :class:`bool`
            indicating compatibility, and the second element is
            a :class:`fontParts.base.compatibility.ComponentCompatibilityReporter`
            instance.

        Example::

            >>> compatible, report = self.isCompatible(otherComponent)
            >>> compatible
            True
            >>> compatible
            [Warning] Component: "A" + "B"
            [Warning] Component: "A" has name A | "B" has name B

        """
        return super().isCompatible(other, BaseComponent)

    def _isCompatible(
        self, other: BaseComponent, reporter: ComponentCompatibilityReporter
    ) -> None:
        """Evaluate interpolation compatibility with another native component.

        This is the environment implementation of :meth:`BaseComponent.isCompatible`.

        :param other: The other :class:`BaseComponent` instance to check
            compatibility with.
        :param reporter: An object used to report compatibility issues.

        .. note::

            Subclasses may override this method.

        """
        component1 = self
        component2 = other
        # base glyphs
        if component1.baseGlyph != component2.baseGlyph:
            reporter.baseDifference = True
            reporter.warning = True

    # ------------
    # Data Queries
    # ------------

    def pointInside(self, point: CoordinateLike) -> bool:
        """Check if `point` lies inside the filled area of the component.

        :param point: The point to check as a :ref:`type-coordinate`.
        :return: :obj:`True` if `point` is inside the filled area of the
            glyph, :obj:`False` otherwise.

        Example::

            >>> glyph.pointInside((40, 65))
            True

        """
        point = normalizers.normalizeCoordinateTuple(point)
        return self._pointInside(point)

    def _pointInside(self, point: CoordinateLike) -> bool:
        """Check if `point` lies inside the filled area of the native component.

        This is the environment implementation of :meth:`BaseComponent.pointInside`.

        :param point: The point to check as a :ref:`type-coordinate`. The value will
            have been normalized with :func:`normalizers.normalizeCoordinateTuple`.
        :return: :class:`bool`.

        .. note::

            Subclasses may override this method.

        """
        pen = PointInsidePen(glyphSet=self.layer, testPoint=point, evenOdd=False)
        self.draw(pen)
        return pen.getResult()

    bounds: dynamicProperty = dynamicProperty(
        "base_bounds",
        """Get the bounds of the component.

        This property is read-only.

        :return: The :class:`Bounds` instance representing the the component's bounds,
            or :obj:`None` if the component is empty.

        Example::

            >>> component.bounds
            Bounds(xMin=10, yMin=30, xMax=765, yMax=643)

        """,
    )

    def _get_base_bounds(self) -> Bounds | None:
        value = self._get_bounds()
        if value is None:
            return None
        value = normalizers.normalizeBoundingBox(value)
        return Bounds(value)

    def _get_bounds(self) -> BoundingBox | None:
        """Get the bounds of the native component.

         This is the environment implementation of the :attr:`BaseComponent.bounds`
         property getter.

        :return: A :class:`tuple` of four :class:`int` or :class:`float` values
             in the form ``(x minimum, y minimum, x maximum, y maximum)``
             representing the bounds of the component, or :obj:`None` if the
             component is empty. The value will be normalized
             with :func:`normalizers.normalizeBoundingBox`.

         .. note::

             Subclasses may override this method.

        """
        pen = BoundsPen(self.layer)
        self.draw(pen)
        return pen.bounds
